@@ -9,6 +9,7 @@
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/prim.h>
 #include <osmocom/core/socket.h>
+#include <osmocom/core/rate_ctr.h>
 
 extern struct llist_head osmo_ss7_instances;
 
@@ -114,6 +115,7 @@ int osmo_ss7_instance_set_pc_fmt(struct osmo_ss7_instance *inst,
 struct osmo_sccp_instance *osmo_ss7_ensure_sccp(struct osmo_ss7_instance *inst);
 
 uint8_t osmo_ss7_pc_width(const struct osmo_ss7_pc_fmt *pc_fmt);
+uint32_t osmo_ss7_pc_normalize(const struct osmo_ss7_pc_fmt *pc_fmt, uint32_t pc);
 
 /***********************************************************************
  * MTP Users (Users of MTP, such as SCCP or ISUP)
@@ -309,6 +311,9 @@ struct osmo_ss7_as {
 	/*! Were we allocated by "simple client" support? */
 	bool simple_client_allocated;
 
+	/*! Rate Counter Group */
+	struct rate_ctr_group *ctrg;
+
 	struct {
 		char *name;
 		char *description;
@@ -347,6 +352,7 @@ int osmo_ss7_as_del_asp(struct osmo_ss7_as *as, const char *asp_name);
 void osmo_ss7_as_destroy(struct osmo_ss7_as *as);
 bool osmo_ss7_as_has_asp(const struct osmo_ss7_as *as,
 			 const struct osmo_ss7_asp *asp);
+bool osmo_ss7_as_down(const struct osmo_ss7_as *as);
 bool osmo_ss7_as_active(const struct osmo_ss7_as *as);
 bool osmo_ss7_as_tmode_compatible_xua(struct osmo_ss7_as *as, uint32_t m3ua_tmt);
 void osmo_ss7_asp_disconnect(struct osmo_ss7_asp *asp);
@@ -360,6 +366,8 @@ struct osmo_ss7_asp_peer {
 	char *host[OSMO_SOCK_MAX_ADDRS];
 	size_t host_cnt;
 	uint16_t port;
+	/* index in "hosts" array marking the SCTP Primary Address, -1 if no explicit Primary Address set */
+	int idx_primary;
 };
 
 enum osmo_ss7_asp_admin_state {
@@ -411,6 +419,9 @@ struct osmo_ss7_asp {
 	/*! Were we allocated by "simple client" support? */
 	bool simple_client_allocated;
 
+	/*! Rate Counter Group */
+	struct rate_ctr_group *ctrg;
+
 	/*! Pending message for non-blocking IPA read */
 	struct msgb *pending_msg;
 
@@ -422,17 +433,44 @@ struct osmo_ss7_asp {
 		bool is_server;
 		enum osmo_ss7_asp_role role;
 		bool role_set_by_vty;
+		bool sctp_role_set_by_vty;
 
 		struct osmo_ss7_asp_peer local;
 		struct osmo_ss7_asp_peer remote;
 		uint8_t qos_class;
+		uint32_t quirks;
+
+		/* T_defs used by the default_lm: */
+		struct osmo_tdef *T_defs_lm;
+
+		struct {
+			bool num_ostreams_present;
+			bool max_instreams_present;
+			bool max_attempts_present;
+			bool max_init_timeo_present;
+			uint16_t num_ostreams_value;
+			uint16_t max_instreams_value;
+			uint16_t max_attempts_value;
+			uint16_t max_init_timeo_value; /* ms */
+		} sctp_init;
 	} cfg;
 };
 
+/*! Peer SG doesn't send NTFY(AS-INACTIVE) after ASP-UP procedure */
+#define OSMO_SS7_ASP_QUIRK_NO_NOTIFY		0x00000001
+/*! Accept DAUD in ASP role (RFC states only permitted in ASP->SG role) */
+#define OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP		0x00000002
+/*! Accept SSNM even if ASP is in AS-INACTIVE state */
+#define OSMO_SS7_ASP_QUIRK_SNM_INACTIVE		0x00000004
+
 int osmo_ss7_asp_peer_snprintf(char* buf, size_t buf_len, struct osmo_ss7_asp_peer *peer);
+void osmo_ss7_asp_peer_init(struct osmo_ss7_asp_peer *peer);
 int osmo_ss7_asp_peer_set_hosts(struct osmo_ss7_asp_peer *peer, void *talloc_ctx,
-				const char* const* hosts, size_t host_cnt);
+				const char *const*hosts, size_t host_cnt);
+int osmo_ss7_asp_peer_set_hosts2(struct osmo_ss7_asp_peer *peer, void *talloc_ctx,
+				const char *const*hosts, size_t host_cnt, int idx_primary);
 int osmo_ss7_asp_peer_add_host(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *host);
+int osmo_ss7_asp_peer_add_host2(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *host, bool is_primary_addr);
 
 struct osmo_ss7_asp *
 osmo_ss7_asp_find_by_name(struct osmo_ss7_instance *inst, const char *name);
@@ -453,6 +491,8 @@ int osmo_ss7_asp_restart(struct osmo_ss7_asp *asp);
 int osmo_ss7_asp_use_default_lm(struct osmo_ss7_asp *asp, int log_level);
 bool osmo_ss7_asp_active(const struct osmo_ss7_asp *asp);
 int osmo_ss7_asp_get_log_subsys(const struct osmo_ss7_asp *asp);
+const char *osmo_ss7_asp_get_name(const struct osmo_ss7_asp *asp);
+enum osmo_ss7_asp_protocol osmo_ss7_asp_get_proto(const struct osmo_ss7_asp *asp);
 
 /*! Weak function to handle payload for unknown/unsupported PPID or IPA StreamID.
  *  This function can be overridden by application code to implement whatever handling
@@ -495,6 +535,12 @@ struct osmo_xua_server {
 		bool accept_dyn_reg;
 		struct osmo_ss7_asp_peer local;
 		enum osmo_ss7_asp_protocol proto;
+		struct {
+			bool num_ostreams_present;
+			bool max_instreams_present;
+			uint16_t num_ostreams_value;
+			uint16_t max_instreams_value;
+		} sctp_init;
 	} cfg;
 };
 
@@ -548,6 +594,8 @@ osmo_sccp_simple_server_add_clnt(struct osmo_sccp_instance *inst,
 				 int local_port, int remote_port,
 				 const char *remote_ip);
 
+void osmo_sccp_set_max_optional_data(struct osmo_sccp_instance *inst, int val);
+
 enum osmo_ss7_as_traffic_mode osmo_ss7_tmode_from_xua(uint32_t in);
 int osmo_ss7_tmode_to_xua(enum osmo_ss7_as_traffic_mode tmod);
 
