@@ -127,32 +127,7 @@ const struct value_string osmo_ss7_asp_role_names[] = {
 	{ 0, NULL }
 };
 
-/* check if the given transport and ASP protocols are compatible (and implemented) */
-bool ss7_asp_protocol_check_trans_proto(enum osmo_ss7_asp_protocol proto, int trans_proto)
-{
-	switch (proto) {
-	case OSMO_SS7_ASP_PROT_IPA:
-		if (trans_proto == IPPROTO_TCP)
-			return true;
-		return false;
-	case OSMO_SS7_ASP_PROT_SUA:
-		if (trans_proto == IPPROTO_SCTP)
-			return true;
-		return false;
-	case OSMO_SS7_ASP_PROT_M3UA:
-		if (trans_proto == IPPROTO_SCTP)
-			return true;
-		if (trans_proto == IPPROTO_TCP)
-			return true;
-		return false;
-	case OSMO_SS7_ASP_PROT_NONE:
-	default:
-		return false;
-	}
-}
-
-/* get _default_ transport protocol for the given ASP protocol */
-int ss7_default_trans_proto_for_asp_proto(enum osmo_ss7_asp_protocol proto)
+int ss7_asp_proto_to_ip_proto(enum osmo_ss7_asp_protocol proto)
 {
 	switch (proto) {
 	case OSMO_SS7_ASP_PROT_IPA:
@@ -193,56 +168,188 @@ static const struct rate_ctr_group_desc ss7_asp_rcgd = {
 };
 static unsigned int g_ss7_asp_rcg_idx;
 
-int ss7_asp_apply_new_local_address(const struct osmo_ss7_asp *asp, unsigned int loc_idx)
+void osmo_ss7_asp_peer_init(struct osmo_ss7_asp_peer *peer)
 {
-	const char *new_loc_addr;
-	int fd;
-
-	OSMO_ASSERT(loc_idx < asp->cfg.local.host_cnt);
-	new_loc_addr = asp->cfg.local.host[loc_idx];
-
-	LOGPASP(asp, DLSS7, LOGL_INFO, "Add local address %s\n",
-		new_loc_addr);
-
-	if (asp->cfg.is_server)
-		fd = osmo_stream_srv_get_fd(asp->server);
-	else
-		fd = osmo_stream_cli_get_fd(asp->client);
-
-	if (fd < 0)
-		return fd;
-
-	return osmo_sock_multiaddr_add_local_addr(fd, &new_loc_addr, 1);
+	memset(peer, 0, sizeof(*peer));
+	peer->idx_primary = -1;
 }
 
-int ss7_asp_apply_drop_local_address(const struct osmo_ss7_asp *asp, unsigned int loc_idx)
+int osmo_ss7_asp_peer_snprintf(char *buf, size_t buf_len, struct osmo_ss7_asp_peer *peer)
 {
-	const char *new_loc_addr;
-	int fd;
+	int len = 0, offset = 0, rem = buf_len;
+	int ret, i;
+	char *after;
+	char *primary;
 
-	OSMO_ASSERT(loc_idx < asp->cfg.local.host_cnt);
-	new_loc_addr = asp->cfg.local.host[loc_idx];
+	if (buf_len < 3)
+		return -EINVAL;
 
-	LOGPASP(asp, DLSS7, LOGL_INFO, "Remove local address %s\n",
-		new_loc_addr);
+	if (peer->host_cnt > 1) {
+		ret = snprintf(buf, rem, "(");
+		if (ret < 0)
+			return ret;
+		OSMO_SNPRINTF_RET(ret, rem, offset, len);
+	}
+	for (i = 0; i < peer->host_cnt; i++) {
+		primary = (peer->idx_primary == i) ? "*" : "";
+		if (peer->host_cnt == 1)
+			after = "";
+		else
+			after = (i == (peer->host_cnt - 1)) ? ")" : "|";
+		ret = snprintf(buf + offset, rem, "%s%s%s", peer->host[i] ? : "0.0.0.0", primary, after);
+		OSMO_SNPRINTF_RET(ret, rem, offset, len);
+	}
+	ret = snprintf(buf + offset, rem, ":%u", peer->port);
+	if (ret < 0)
+		return ret;
+	OSMO_SNPRINTF_RET(ret, rem, offset, len);
 
-	if (asp->cfg.is_server)
-		fd = osmo_stream_srv_get_fd(asp->server);
-	else
-		fd = osmo_stream_cli_get_fd(asp->client);
+	return len;
+}
 
-	if (fd < 0)
-		return fd;
+/*! \brief Set (copy) addresses for a given ASP peer. Previous addresses are freed.
+ *  \param[in] peer Application Server Process peer whose addresses are to be set.
+ *  \param[in] talloc_ctx talloc context used to allocate new addresses.
+ *  \param[in] hosts Array of strings containing IP addresses.
+ *  \param[in] host_cnt Number of strings in hosts
+ *  \param[in] idx_primary Index in "hosts" array marking the SCTP Primary Address, -1 if no explicit Primary Address set
+ *  \returns 0 on success; negative otherwise */
+int osmo_ss7_asp_peer_set_hosts2(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *const*hosts, size_t host_cnt, int idx_primary)
+{
+	int i = 0;
 
-	return osmo_sock_multiaddr_del_local_addr(fd, &new_loc_addr, 1);
+	if (idx_primary >= (int)host_cnt || idx_primary < -1)
+		return -EINVAL;
+
+	if (host_cnt > ARRAY_SIZE(peer->host))
+		return -EINVAL;
+
+	for (; i < host_cnt; i++)
+		osmo_talloc_replace_string(talloc_ctx, &peer->host[i], hosts[i]);
+	for (; i < peer->host_cnt; i++) {
+		talloc_free(peer->host[i]);
+		peer->host[i] = NULL;
+	}
+
+	peer->host_cnt = host_cnt;
+	peer->idx_primary = idx_primary;
+	return 0;
+}
+
+/*! \brief Set (copy) addresses for a given ASP peer. Previous addresses are freed.
+ *  \param[in] peer Application Server Process peer whose addresses are to be set.
+ *  \param[in] talloc_ctx talloc context used to allocate new addresses.
+ *  \param[in] hosts Array of strings containing IP addresses.
+ *  \param[in] host_cnt Number of strings in hosts
+ *  \returns 0 on success; negative otherwise */
+int osmo_ss7_asp_peer_set_hosts(struct osmo_ss7_asp_peer *peer, void *talloc_ctx, const char *const*hosts, size_t host_cnt)
+{
+	return osmo_ss7_asp_peer_set_hosts2(peer, talloc_ctx, hosts, host_cnt, -1);
+}
+
+/* Is string formatted IPv4/v6 addr considered IN(6)ADDR_ANY? */
+static inline bool host_is_ip_anyaddr(const char *host, bool is_v6)
+{
+	/* NULL addr is resolved as 0.0.0.0 (IPv4) by getaddrinfo(), most
+	 * probably for backward-compatibility reasons.
+	 */
+	return is_v6 ? (host && !strcmp(host, "::"))
+		     : (!host || !strcmp(host, "0.0.0.0"));
+}
+
+/*! \brief Append (copy) address to a given ASP peer. Previous addresses are kept.
+ *  \param[in] peer Application Server Process peer the address is appended to.
+ *  \param[in] talloc_ctx talloc context used to allocate new address.
+ *  \param[in] host string containing an IP address.
+ *  \param[in] is_primary_addr whether this IP address is to be added as SCTP Primary Address
+ *  \returns 0 on success; negative otherwise */
+int osmo_ss7_asp_peer_add_host2(struct osmo_ss7_asp_peer *peer, void *talloc_ctx,
+			       const char *host, bool is_primary_addr)
+{
+	int i;
+	bool new_is_v6 = osmo_ip_str_type(host) == AF_INET6;
+	bool new_is_any = host_is_ip_anyaddr(host, new_is_v6);
+	struct osmo_sockaddr_str addr_str;
+
+	if (osmo_sockaddr_str_from_str(&addr_str, host, 0) < 0)
+		return -EINVAL;
+
+	if (new_is_any) {
+		/* Makes no sense to have INET(6)_ANY many times, or INET(6)_ANY
+		 * together with specific addresses, be it of same or different
+		 * IP version: */
+		if (peer->host_cnt != 0)
+			return -EINVAL;
+
+		/* Makes no sense to have INET(6)_ANY as primary: */
+		if (is_primary_addr)
+			return -EINVAL;
+
+		if (peer->host_cnt >= ARRAY_SIZE(peer->host))
+			return -EINVAL;
+		osmo_talloc_replace_string(talloc_ctx, &peer->host[peer->host_cnt], host);
+		peer->host_cnt++;
+		return 0;
+	}
+
+	/* Makes no sense to add specific address to set if INET(6)_ANY
+	 * is already set, be it from same or different IP version: */
+	for (i = 0; i < peer->host_cnt; i++) {
+		bool iter_is_v6 = osmo_ip_str_type(peer->host[i]) == AF_INET6;
+		if (host_is_ip_anyaddr(peer->host[i], iter_is_v6))
+			return -EINVAL;
+	}
+	/* Reached this point, no INET(6)_ANY address is set nor we are adding an INET(6)_ANY address. */
+
+	/* Check if address already exists, and then if primary flags need to be changed: */
+	for (i = 0; i < peer->host_cnt; i++) {
+		struct osmo_sockaddr_str it_addr_str;
+		bool it_is_primary;
+		osmo_sockaddr_str_from_str(&it_addr_str, peer->host[i], 0);
+
+		if (osmo_sockaddr_str_cmp(&addr_str, &it_addr_str) != 0)
+			continue;
+		it_is_primary = (peer->idx_primary == i);
+		if (is_primary_addr == it_is_primary) {
+			/* Nothing to do, return below */
+		} else if (is_primary_addr && !it_is_primary) {
+			/* Mark it as primary: */
+			peer->idx_primary = i;
+		} else { /* if (!is_primary_addr && it_is_primary) { */
+			/* mark it as non-primary: */
+			peer->idx_primary = -1;
+		}
+		return 0;
+	}
+
+	if (peer->host_cnt >= ARRAY_SIZE(peer->host))
+		return -EINVAL;
+
+	osmo_talloc_replace_string(talloc_ctx, &peer->host[peer->host_cnt], host);
+	if (is_primary_addr)
+		peer->idx_primary = peer->host_cnt;
+	peer->host_cnt++;
+	return 0;
+}
+
+/*! \brief Append (copy) address to a given ASP peer. Previous addresses are kept.
+ *  \param[in] peer Application Server Process peer the address is appended to.
+ *  \param[in] talloc_ctx talloc context used to allocate new address.
+ *  \param[in] host string containing an IP address.
+ *  \returns 0 on success; negative otherwise */
+int osmo_ss7_asp_peer_add_host(struct osmo_ss7_asp_peer *peer, void *talloc_ctx,
+			       const char *host)
+{
+	return osmo_ss7_asp_peer_add_host2(peer, talloc_ctx, host, false);
 }
 
 int ss7_asp_apply_peer_primary_address(const struct osmo_ss7_asp *asp)
 {
+	struct osmo_fd *ofd;
 	struct osmo_sockaddr_str addr_str;
 	struct osmo_sockaddr addr;
 	uint16_t local_port;
-	int fd, rc;
+	int rc;
 
 	/* No SCTP Peer Primary Address explicitly configured, do nothing. */
 	if (asp->cfg.local.idx_primary == -1)
@@ -250,16 +357,13 @@ int ss7_asp_apply_peer_primary_address(const struct osmo_ss7_asp *asp)
 	OSMO_ASSERT(asp->cfg.local.idx_primary < asp->cfg.local.host_cnt);
 
 	if (asp->cfg.is_server)
-		fd = osmo_stream_srv_get_fd(asp->server);
+		ofd = osmo_stream_srv_get_ofd(asp->server);
 	else
-		fd = osmo_stream_cli_get_fd(asp->client);
-
-	if (fd < 0)
-		return fd;
+		ofd = osmo_stream_cli_get_ofd(asp->client);
 
 	if (asp->cfg.local.port == 0) {
 		char port_buf[16];
-		osmo_sock_get_local_ip_port(fd, port_buf, sizeof(port_buf));
+		osmo_sock_get_local_ip_port(ofd->fd, port_buf, sizeof(port_buf));
 		local_port = atoi(port_buf);
 	} else {
 		local_port = asp->cfg.local.port;
@@ -274,16 +378,17 @@ int ss7_asp_apply_peer_primary_address(const struct osmo_ss7_asp *asp)
 		return rc;
 	LOGPASP(asp, DLSS7, LOGL_INFO, "Set Peer's Primary Address %s\n",
 		osmo_sockaddr_to_str(&addr));
-	rc = _setsockopt_peer_primary_addr(fd, &addr);
+	rc = _setsockopt_peer_primary_addr(ofd->fd, &addr);
 
 	return rc;
 }
 
 int ss7_asp_apply_primary_address(const struct osmo_ss7_asp *asp)
 {
+	struct osmo_fd *ofd;
 	struct osmo_sockaddr_str addr_str;
 	struct osmo_sockaddr addr;
-	int fd, rc;
+	int rc;
 
 	/* No SCTP Primary Address explicitly configured, do nothing. */
 	if (asp->cfg.remote.idx_primary == -1)
@@ -291,12 +396,9 @@ int ss7_asp_apply_primary_address(const struct osmo_ss7_asp *asp)
 	OSMO_ASSERT(asp->cfg.remote.idx_primary < asp->cfg.remote.host_cnt);
 
 	if (asp->cfg.is_server)
-		fd = osmo_stream_srv_get_fd(asp->server);
+		ofd = osmo_stream_srv_get_ofd(asp->server);
 	else
-		fd = osmo_stream_cli_get_fd(asp->client);
-
-	if (fd < 0)
-		return fd;
+		ofd = osmo_stream_cli_get_ofd(asp->client);
 
 	rc = osmo_sockaddr_str_from_str(&addr_str,
 					asp->cfg.remote.host[asp->cfg.remote.idx_primary],
@@ -308,7 +410,7 @@ int ss7_asp_apply_primary_address(const struct osmo_ss7_asp *asp)
 		return rc;
 	LOGPASP(asp, DLSS7, LOGL_INFO, "Set Primary Address %s\n",
 		osmo_sockaddr_to_str(&addr));
-	rc = _setsockopt_primary_addr(fd, &addr);
+	rc = _setsockopt_primary_addr(ofd->fd, &addr);
 	return rc;
 }
 
@@ -458,7 +560,7 @@ static void chop_v4_mapped_on_v6_prefix(char *buf)
  *  \param[in] fd socket descriptor of given socket
  *  \returns SS7 ASP in case a matching one is found; NULL otherwise */
 struct osmo_ss7_asp *
-ss7_asp_find_by_socket_addr(int fd, int trans_proto)
+ss7_asp_find_by_socket_addr(int fd)
 {
 	struct osmo_ss7_instance *inst;
 	struct sockaddr_storage sa_l, sa_r;
@@ -468,6 +570,7 @@ ss7_asp_find_by_socket_addr(int fd, int trans_proto)
 	uint16_t local_port, remote_port;
 	bool loc_is_v6, rem_is_v6;
 	int rc;
+	int i;
 
 	OSMO_ASSERT(ss7_initialized);
 	/* convert local and remote IP to string */
@@ -505,19 +608,39 @@ ss7_asp_find_by_socket_addr(int fd, int trans_proto)
 	llist_for_each_entry(inst, &osmo_ss7_instances, list) {
 		struct osmo_ss7_asp *asp;
 		llist_for_each_entry(asp, &inst->asp_list, list) {
-			if (asp->cfg.trans_proto != trans_proto)
-				continue;
 			if (asp->cfg.local.port != local_port)
 				continue;
 			if (asp->cfg.remote.port && asp->cfg.remote.port != remote_port)
 				continue;
 
-			if (!ss7_asp_peer_match_host(&asp->cfg.local, hostbuf_l, loc_is_v6))
+			for (i = 0; i < asp->cfg.local.host_cnt; i++) {
+				bool iter_is_v6 = osmo_ip_str_type(asp->cfg.local.host[i]) == AF_INET6;
+				bool iter_is_anyaddr = host_is_ip_anyaddr(asp->cfg.local.host[i], iter_is_v6);
+				/* "::" (v6) covers "0.0.0.0" (v4), but not otherwise */
+				if (iter_is_v6 != loc_is_v6 &&
+				    !(iter_is_v6 && iter_is_anyaddr))
+					continue;
+				if (iter_is_anyaddr ||
+				    !strcmp(asp->cfg.local.host[i], hostbuf_l))
+					break;
+			}
+			if (i == asp->cfg.local.host_cnt)
 				continue; /* didn't match any local.host */
 
 			/* If no remote host was set, it's probably a server and hence we match any cli src */
 			if (asp->cfg.remote.host_cnt) {
-				if (!ss7_asp_peer_match_host(&asp->cfg.remote, hostbuf_r, rem_is_v6))
+				for (i = 0; i < asp->cfg.remote.host_cnt; i++) {
+					bool iter_is_v6 = osmo_ip_str_type(asp->cfg.remote.host[i]) == AF_INET6;
+					bool iter_is_anyaddr = host_is_ip_anyaddr(asp->cfg.remote.host[i], iter_is_v6);
+					/* "::" (v6) covers "0.0.0.0" (v4), but not otherwise */
+					if (iter_is_v6 != rem_is_v6 &&
+					    !(iter_is_v6 && iter_is_anyaddr))
+						continue;
+					if (iter_is_anyaddr ||
+					    !strcmp(asp->cfg.remote.host[i], hostbuf_r))
+						break;
+				}
+				if (i == asp->cfg.remote.host_cnt)
 					continue; /* didn't match any remote.host */
 			}
 
@@ -530,18 +653,9 @@ ss7_asp_find_by_socket_addr(int fd, int trans_proto)
 
 struct osmo_ss7_asp *ss7_asp_alloc(struct osmo_ss7_instance *inst, const char *name,
 				   uint16_t remote_port, uint16_t local_port,
-				   int trans_proto, enum osmo_ss7_asp_protocol proto)
+				   enum osmo_ss7_asp_protocol proto)
 {
-	struct osmo_ss7_asp *asp;
-
-	if (!ss7_asp_protocol_check_trans_proto(proto, trans_proto)) {
-		LOGP(DLSCCP, LOGL_ERROR,
-		     "ASP protocol '%s' with transport protocol %d is not supported",
-		     osmo_ss7_asp_protocol_name(proto), trans_proto);
-		return NULL;
-	}
-
-	asp = talloc_zero(inst, struct osmo_ss7_asp);
+	struct osmo_ss7_asp *asp = talloc_zero(inst, struct osmo_ss7_asp);
 	asp->ctrg = rate_ctr_group_alloc(asp, &ss7_asp_rcgd, g_ss7_asp_rcg_idx++);
 	if (!asp->ctrg) {
 		talloc_free(asp);
@@ -553,7 +667,6 @@ struct osmo_ss7_asp *ss7_asp_alloc(struct osmo_ss7_instance *inst, const char *n
 	asp->cfg.remote.port = remote_port;
 	osmo_ss7_asp_peer_init(&asp->cfg.local);
 	asp->cfg.local.port = local_port;
-	asp->cfg.trans_proto = trans_proto;
 	asp->cfg.proto = proto;
 	asp->cfg.name = talloc_strdup(asp, name);
 
@@ -603,7 +716,6 @@ void osmo_ss7_asp_destroy(struct osmo_ss7_asp *asp)
 
 static int xua_cli_read_cb(struct osmo_stream_cli *conn);
 static int ipa_cli_read_cb(struct osmo_stream_cli *conn);
-static int m3ua_tcp_cli_read_cb(struct osmo_stream_cli *conn);
 static int xua_cli_connect_cb(struct osmo_stream_cli *cli);
 
 int osmo_ss7_asp_restart(struct osmo_ss7_asp *asp)
@@ -639,27 +751,13 @@ int osmo_ss7_asp_restart(struct osmo_ss7_asp *asp)
 		osmo_stream_cli_set_port(asp->client, asp->cfg.remote.port);
 		osmo_stream_cli_set_local_addrs(asp->client, (const char **)asp->cfg.local.host, asp->cfg.local.host_cnt);
 		osmo_stream_cli_set_local_port(asp->client, asp->cfg.local.port);
-		osmo_stream_cli_set_proto(asp->client, asp->cfg.trans_proto);
+		osmo_stream_cli_set_proto(asp->client, ss7_asp_proto_to_ip_proto(asp->cfg.proto));
 		osmo_stream_cli_set_reconnect_timeout(asp->client, 5);
 		osmo_stream_cli_set_connect_cb(asp->client, xua_cli_connect_cb);
-		switch (asp->cfg.proto) {
-		case OSMO_SS7_ASP_PROT_IPA:
-			OSMO_ASSERT(asp->cfg.trans_proto == IPPROTO_TCP);
+		if (asp->cfg.proto == OSMO_SS7_ASP_PROT_IPA)
 			osmo_stream_cli_set_read_cb(asp->client, ipa_cli_read_cb);
-			break;
-		case OSMO_SS7_ASP_PROT_M3UA:
-			if (asp->cfg.trans_proto == IPPROTO_SCTP)
-				osmo_stream_cli_set_read_cb(asp->client, xua_cli_read_cb);
-			else if (asp->cfg.trans_proto == IPPROTO_TCP)
-				osmo_stream_cli_set_read_cb(asp->client, m3ua_tcp_cli_read_cb);
-			else
-				OSMO_ASSERT(0);
-			break;
-		default:
-			OSMO_ASSERT(asp->cfg.trans_proto == IPPROTO_SCTP);
+		else
 			osmo_stream_cli_set_read_cb(asp->client, xua_cli_read_cb);
-			break;
-		}
 		osmo_stream_cli_set_data(asp->client, asp);
 		byte = 1; /*AUTH is needed by ASCONF. enable, don't abort socket creation if AUTH can't be enabled */
 		osmo_stream_cli_set_param(asp->client, OSMO_STREAM_CLI_PAR_SCTP_SOCKOPT_AUTH_SUPPORTED, &byte, sizeof(byte));
@@ -717,14 +815,6 @@ bool osmo_ss7_asp_active(const struct osmo_ss7_asp *asp)
 	return asp->fi->state == XUA_ASP_S_ACTIVE;
 }
 
-bool ss7_asp_is_started(const struct osmo_ss7_asp *asp)
-{
-	if (asp->cfg.is_server)
-		return !!asp->server;
-	else
-		return !!asp->client;
-}
-
 /***********************************************************************
  * libosmo-netif integration for SCTP stream server/client
  ***********************************************************************/
@@ -791,15 +881,13 @@ static void log_sctp_notification(struct osmo_ss7_asp *asp, const char *pfx,
 /* netif code tells us we can read something from the socket */
 int ss7_asp_ipa_srv_conn_cb(struct osmo_stream_srv *conn)
 {
-	int fd = osmo_stream_srv_get_fd(conn);
+	struct osmo_fd *ofd = osmo_stream_srv_get_ofd(conn);
 	struct osmo_ss7_asp *asp = osmo_stream_srv_get_data(conn);
 	struct msgb *msg = NULL;
 	int rc;
 
-	OSMO_ASSERT(fd >= 0);
-
 	/* read IPA message from socket and process it */
-	rc = ipa_msg_recv_buffered(fd, &msg, &asp->pending_msg);
+	rc = ipa_msg_recv_buffered(ofd->fd, &msg, &asp->pending_msg);
 	LOGPASP(asp, DLSS7, LOGL_DEBUG, "%s(): ipa_msg_recv_buffered() returned %d\n",
 		__func__, rc);
 	if (rc <= 0) {
@@ -818,9 +906,7 @@ int ss7_asp_ipa_srv_conn_cb(struct osmo_stream_srv *conn)
 	}
 	msg->dst = asp;
 	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
-	/* we can use the 'fd' return value of osmo_stream_srv_get_fd() here unverified as all we do
-	 * is 'roll the dice' to obtain a 4-bit SLS value. */
-	return ipa_rx_msg(asp, msg, fd & 0xf);
+	return ipa_rx_msg(asp, msg, ofd->fd & 0xf);
 }
 
 /* netif code tells us we can read something from the socket */
@@ -890,91 +976,16 @@ out:
 	return rc;
 }
 
-/* netif code tells us we can read something from the socket */
-int ss7_asp_m3ua_tcp_srv_conn_cb(struct osmo_stream_srv *conn)
-{
-	struct osmo_ss7_asp *asp = osmo_stream_srv_get_data(conn);
-	int fd = osmo_stream_srv_get_fd(conn);
-	struct msgb *msg = asp->pending_msg;
-	const struct xua_common_hdr *hdr;
-	size_t msg_length;
-	int rc;
-
-	OSMO_ASSERT(fd >= 0);
-
-	if (msg == NULL) {
-		msg = m3ua_msgb_alloc(__func__);
-		asp->pending_msg = msg;
-	}
-
-	/* read message header first */
-	if (msg->len < sizeof(*hdr)) {
-		errno = 0;
-		rc = recv(fd, msg->tail, sizeof(*hdr) - msg->len, 0);
-		if (rc <= 0) {
-			if (errno == EAGAIN || errno == EINTR)
-				return 0; /* need more data */
-			osmo_stream_srv_destroy(conn);
-			asp->pending_msg = NULL;
-			msgb_free(msg);
-			return rc;
-		}
-
-		msgb_put(msg, rc);
-		if (msg->len < sizeof(*hdr))
-			return 0; /* need more data */
-	}
-
-	hdr = (const struct xua_common_hdr *)msg->data;
-	msg_length = ntohl(hdr->msg_length); /* includes sizeof(*hdr) */
-
-	/* read the rest of the message */
-	if (msg->len < msg_length) {
-		errno = 0;
-		rc = recv(fd, msg->tail, msg_length - msg->len, 0);
-		if (rc <= 0) {
-			if (errno == EAGAIN || errno == EINTR)
-				return 0; /* need more data */
-			osmo_stream_srv_destroy(conn);
-			asp->pending_msg = NULL;
-			msgb_free(msg);
-			return rc;
-		}
-
-		msgb_put(msg, rc);
-		if (msg->len < msg_length)
-			return 0; /* need more data */
-	}
-
-	msg->dst = asp;
-	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
-
-	/* spoof SCTP Stream ID */
-	if (hdr->msg_class == M3UA_MSGC_XFER)
-		msgb_sctp_stream(msg) = 1;
-	else
-		msgb_sctp_stream(msg) = 0;
-
-	rc = m3ua_rx_msg(asp, msg);
-	asp->pending_msg = NULL;
-	msgb_free(msg);
-
-	return rc;
-}
-
 /* client has established SCTP connection to server */
 static int xua_cli_connect_cb(struct osmo_stream_cli *cli)
 {
-	int fd = osmo_stream_cli_get_fd(cli);
+	struct osmo_fd *ofd = osmo_stream_cli_get_ofd(cli);
 	struct osmo_ss7_asp *asp = osmo_stream_cli_get_data(cli);
 	int rc = 0;
 
-	if (fd < 0)
-		return fd;
-
 	/* update the socket name */
 	talloc_free(asp->sock_name);
-	asp->sock_name = osmo_sock_get_name(asp, fd);
+	asp->sock_name = osmo_sock_get_name(asp, ofd->fd);
 
 	LOGPASP(asp, DLSS7, LOGL_INFO, "Client connected %s\n", asp->sock_name);
 
@@ -982,7 +993,7 @@ static int xua_cli_connect_cb(struct osmo_stream_cli *cli)
 	 * fed and the local port is known for sure. Apply SCTP Primary addresses
 	 * if needed:
 	 */
-	if (asp->cfg.trans_proto == IPPROTO_SCTP) {
+	if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA) {
 		rc = ss7_asp_apply_peer_primary_address(asp);
 		rc = ss7_asp_apply_primary_address(asp);
 	}
@@ -992,7 +1003,7 @@ static int xua_cli_connect_cb(struct osmo_stream_cli *cli)
 		 * established */
 		xua_asp_send_xlm_prim_simple(asp, OSMO_XLM_PRIM_M_SCTP_ESTABLISH, PRIM_OP_INDICATION);
 	} else {
-		/* directly ask the ASP FSM to start by sending an ASP-UP ... */
+		/* directly as the ASP FSM to start by sending an ASP-UP ... */
 		osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
 	}
 
@@ -1018,15 +1029,13 @@ static void xua_cli_close_and_reconnect(struct osmo_stream_cli *cli)
 /* read call-back for IPA/SCCPlite socket */
 static int ipa_cli_read_cb(struct osmo_stream_cli *conn)
 {
-	int fd = osmo_stream_cli_get_fd(conn);
+	struct osmo_fd *ofd = osmo_stream_cli_get_ofd(conn);
 	struct osmo_ss7_asp *asp = osmo_stream_cli_get_data(conn);
 	struct msgb *msg = NULL;
 	int rc;
 
-	OSMO_ASSERT(fd >= 0);
-
 	/* read IPA message from socket and process it */
-	rc = ipa_msg_recv_buffered(fd, &msg, &asp->pending_msg);
+	rc = ipa_msg_recv_buffered(ofd->fd, &msg, &asp->pending_msg);
 	LOGPASP(asp, DLSS7, LOGL_DEBUG, "%s(): ipa_msg_recv_buffered() returned %d\n",
 		__func__, rc);
 	if (rc <= 0) {
@@ -1045,81 +1054,7 @@ static int ipa_cli_read_cb(struct osmo_stream_cli *conn)
 	}
 	msg->dst = asp;
 	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
-	/* we can use the 'fd' return value of osmo_stream_srv_get_fd() here unverified as all we do
-	 * is 'roll the dice' to obtain a 4-bit SLS value. */
-	return ipa_rx_msg(asp, msg, fd & 0xf);
-}
-
-/* read call-back for M3UA-over-TCP socket */
-static int m3ua_tcp_cli_read_cb(struct osmo_stream_cli *conn)
-{
-	struct osmo_ss7_asp *asp = osmo_stream_cli_get_data(conn);
-	int fd = osmo_stream_cli_get_fd(conn);
-	struct msgb *msg = asp->pending_msg;
-	const struct xua_common_hdr *hdr;
-	size_t msg_length;
-	int rc;
-
-	OSMO_ASSERT(fd >= 0);
-
-	if (msg == NULL) {
-		msg = m3ua_msgb_alloc(__func__);
-		asp->pending_msg = msg;
-	}
-
-	/* read message header first */
-	if (msg->len < sizeof(*hdr)) {
-		errno = 0;
-		rc = recv(fd, msg->tail, sizeof(*hdr) - msg->len, 0);
-		if (rc <= 0) {
-			if (errno == EAGAIN || errno == EINTR)
-				return 0; /* need more data */
-			xua_cli_close_and_reconnect(conn);
-			asp->pending_msg = NULL;
-			msgb_free(msg);
-			return rc;
-		}
-
-		msgb_put(msg, rc);
-		if (msg->len < sizeof(*hdr))
-			return 0; /* need more data */
-	}
-
-	hdr = (const struct xua_common_hdr *)msg->data;
-	msg_length = ntohl(hdr->msg_length); /* includes sizeof(*hdr) */
-
-	/* read the rest of the message */
-	if (msg->len < msg_length) {
-		errno = 0;
-		rc = recv(fd, msg->tail, msg_length - msg->len, 0);
-		if (rc <= 0) {
-			if (errno == EAGAIN || errno == EINTR)
-				return 0; /* need more data */
-			xua_cli_close_and_reconnect(conn);
-			asp->pending_msg = NULL;
-			msgb_free(msg);
-			return rc;
-		}
-
-		msgb_put(msg, rc);
-		if (msg->len < msg_length)
-			return 0; /* need more data */
-	}
-
-	msg->dst = asp;
-	rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_TOTAL);
-
-	/* spoof SCTP Stream ID */
-	if (hdr->msg_class == M3UA_MSGC_XFER)
-		msgb_sctp_stream(msg) = 1;
-	else
-		msgb_sctp_stream(msg) = 0;
-
-	rc = m3ua_rx_msg(asp, msg);
-	asp->pending_msg = NULL;
-	msgb_free(msg);
-
-	return rc;
+	return ipa_rx_msg(asp, msg, ofd->fd & 0xf);
 }
 
 static int xua_cli_read_cb(struct osmo_stream_cli *conn)
@@ -1330,28 +1265,3 @@ enum osmo_ss7_asp_protocol osmo_ss7_asp_get_proto(const struct osmo_ss7_asp *asp
 {
 	return asp->cfg.proto;
 }
-
-/*! \brief Get the transport proto of a given ASP
- *  \param[in] asp The ASP for which the transport proto is requested
- *  \returns The transport proto of the ASP (one of IPPROTO_*)
- */
-int osmo_ss7_asp_get_trans_proto(const struct osmo_ss7_asp *asp)
-{
-	return asp->cfg.trans_proto;
-}
-
-/*! \brief Get the fd of a given ASP
- *  \param[in] asp The ASP for which the fd is requested
- *  \returns The fd of the ASP if acailable, negative otherwise
- */
-int ss7_asp_get_fd(const struct osmo_ss7_asp *asp)
-{
-	if (asp->cfg.is_server) {
-		if (asp->server)
-			return osmo_stream_srv_get_fd(asp->server);
-	} else {
-		if (asp->client)
-			return osmo_stream_cli_get_fd(asp->client);
-	}
-	return -1;
-}
