@@ -28,7 +28,6 @@
 
 #include <netdb.h>
 #include <arpa/inet.h>
-#include <sys/ioctl.h>
 
 #include <osmocom/core/sockaddr_str.h>
 
@@ -48,13 +47,6 @@
 #include "sccp_internal.h"
 #include "ss7_internal.h"
 
-#include <netinet/tcp.h>
-
-#ifdef HAVE_LIBSCTP
-#include <netinet/sctp.h>
-#include <osmocom/netif/sctp.h>
-#endif
-
 #define XUA_VAR_STR	"(sua|m3ua|ipa)"
 
 #define XUA_VAR_HELP_STR		\
@@ -62,22 +54,6 @@
 	"MTP3 User Adaptation\n"	\
 	"IPA Multiplex (SCCP Lite)\n"
 
-/* netinet/tcp.h */
-static const struct value_string tcp_info_state_values[] = {
-	{ TCP_ESTABLISHED,	"ESTABLISHED" },
-	{ TCP_SYN_SENT,		"SYN_SENT" },
-	{ TCP_SYN_RECV,		"SYN_RECV" },
-	{ TCP_FIN_WAIT1,	"FIN_WAIT1" },
-	{ TCP_FIN_WAIT2,	"FIN_WAIT2" },
-	{ TCP_TIME_WAIT,	"TIME_WAIT" },
-	{ TCP_CLOSE,		"CLOSE" },
-	{ TCP_CLOSE_WAIT,	"CLOSE_WAIT" },
-	{ TCP_LAST_ACK,		"LAST_ACK" },
-	{ TCP_LISTEN,		"LISTEN" },
-	{ TCP_CLOSING,		"CLOSING" },
-	{}
-};
-
 static const struct value_string asp_quirk_names[] = {
 	{ OSMO_SS7_ASP_QUIRK_NO_NOTIFY,		"no_notify" },
 	{ OSMO_SS7_ASP_QUIRK_DAUD_IN_ASP,	"daud_in_asp" },
@@ -491,7 +467,7 @@ DEFUN_ATTR(cs7_xua, cs7_xua_cmd,
 	if (!xs) {
 		xs = osmo_ss7_xua_server_create(inst, proto, port, NULL);
 		if (!xs)
-			return CMD_WARNING;
+			return CMD_SUCCESS;
 		/* Drop first dummy address created automatically by _create(): */
 		osmo_ss7_xua_server_set_local_hosts(xs, NULL, 0);
 	}
@@ -535,22 +511,6 @@ DEFUN_ATTR(xua_local_ip, xua_local_ip_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN_ATTR(xua_no_local_ip, xua_no_local_ip_cmd,
-	   "no local-ip " VTY_IPV46_CMD,
-	   NO_STR "Configure the Local IP Address for xUA\n"
-	   "IPv4 Address to use for XUA\n"
-	   "IPv6 Address to use for XUA\n",
-	   CMD_ATTR_NODE_EXIT)
-{
-	struct osmo_xua_server *xs = vty->index;
-
-	if (osmo_ss7_xua_server_del_local_host(xs, argv[0]) != 0) {
-		vty_out(vty, "%% Failed deleting local address '%s' from set%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-	return CMD_SUCCESS;
-}
-
 DEFUN_ATTR(xua_accept_dyn_asp, xua_accept_dyn_asp_cmd,
 	   "accept-asp-connections (pre-configured|dynamic-permitted)",
 	   "Define what kind of ASP connections to accept\n"
@@ -634,33 +594,11 @@ static void write_one_xua(struct vty *vty, struct osmo_xua_server *xs)
 
 static void vty_dump_xua_server(struct vty *vty, struct osmo_xua_server *xs)
 {
-	char buf[OSMO_SOCK_MULTIADDR_PEER_STR_MAXLEN];
+	char buf[512];
 	const char *proto = get_value_string(osmo_ss7_asp_protocol_vals, xs->cfg.proto);
-	int fd = xs->server ? osmo_stream_srv_link_get_fd(xs->server) : -1;
-
-	if (fd < 0) {
-		if (osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &xs->cfg.local) < 0)
-			snprintf(buf, sizeof(buf), "<error>");
-	} else {
-		char hostbuf[OSMO_SOCK_MAX_ADDRS][INET6_ADDRSTRLEN];
-		size_t num_hostbuf = ARRAY_SIZE(hostbuf);
-		char portbuf[6];
-		int rc;
-		rc = osmo_sock_multiaddr_get_ip_and_port(fd, ss7_asp_proto_to_ip_proto(xs->cfg.proto),
-							 &hostbuf[0][0], &num_hostbuf, sizeof(hostbuf[0]),
-							 portbuf, sizeof(portbuf), true);
-		if (rc < 0) {
-			snprintf(buf, sizeof(buf), "<error>");
-		} else {
-			if (num_hostbuf > ARRAY_SIZE(hostbuf))
-				num_hostbuf = ARRAY_SIZE(hostbuf);
-			osmo_multiaddr_ip_and_port_snprintf(buf, sizeof(buf),
-							    &hostbuf[0][0], num_hostbuf, sizeof(hostbuf[0]),
-							    portbuf);
-		}
-	}
-	vty_out(vty, "xUA server for %s on %s is %s%s",
-		proto, buf, fd >= 0 ? "listening" : "inactive", VTY_NEWLINE);
+	if (osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &xs->cfg.local) < 0)
+		snprintf(buf, sizeof(buf), "<error>");
+	vty_out(vty, "xUA server for %s on %s%s", proto, buf, VTY_NEWLINE);
 }
 
 DEFUN(show_cs7_xua, show_cs7_xua_cmd,
@@ -698,31 +636,6 @@ DEFUN(show_cs7_config, show_cs7_config_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cs7_asp_disconnect, cs7_asp_disconnect_cmd,
-      "cs7 instance <0-15> asp NAME disconnect",
-      CS7_STR "Instance related commands\n" "SS7 Instance Number\n"
-      "ASP related commands\n" "Name of ASP\n"
-      "Disconnect the ASP (client will reconnect)\n")
-{
-	struct osmo_ss7_instance *inst;
-	struct osmo_ss7_asp *asp;
-
-	inst = osmo_ss7_instance_find(atoi(argv[0]));
-	if (!inst) {
-		vty_out(vty, "unknown instance '%s'%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	asp = osmo_ss7_asp_find_by_name(inst, argv[1]);
-	if (!asp) {
-		vty_out(vty, "unknown ASP '%s'%s", argv[1], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	osmo_ss7_asp_disconnect(asp);
-	return CMD_SUCCESS;
-}
-
 
 /***********************************************************************
  * Application Server Process
@@ -801,76 +714,11 @@ DEFUN_ATTR(asp_local_ip, asp_local_ip_cmd,
 {
 	struct osmo_ss7_asp *asp = vty->index;
 	bool is_primary = argc > 1;
-	int old_idx_primary = asp->cfg.local.idx_primary;
-	int old_host_count = asp->cfg.local.host_cnt;
-	int rc;
 
 	if (osmo_ss7_asp_peer_add_host2(&asp->cfg.local, asp, argv[0], is_primary) != 0) {
 		vty_out(vty, "%% Failed adding host '%s' to set%s", argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
-
-	if (!ss7_asp_is_started(asp))
-		return CMD_SUCCESS;
-	if (asp->cfg.proto == OSMO_SS7_ASP_PROT_IPA)
-		return CMD_SUCCESS;
-	/* The SCTP socket is already created. */
-
-	/* dynamically apply the new address if it was added to the set: */
-	if (asp->cfg.local.host_cnt > old_host_count) {
-		if ((rc = ss7_asp_apply_new_local_address(asp, asp->cfg.local.host_cnt - 1)) < 0) {
-			/* Failed, rollback changes: */
-			TALLOC_FREE(asp->cfg.local.host[asp->cfg.local.host_cnt - 1]);
-			asp->cfg.local.host_cnt--;
-			vty_out(vty, "%% Failed adding new local address '%s'%s", argv[0], VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-		vty_out(vty, "%% Local address '%s' added to the active socket bind set%s", argv[0], VTY_NEWLINE);
-	}
-
-	/* dynamically apply the new primary if it changed: */
-	if (is_primary && asp->cfg.local.idx_primary != old_idx_primary) {
-		if ((rc = ss7_asp_apply_peer_primary_address(asp)) < 0) {
-			/* Failed, rollback changes: */
-			asp->cfg.local.idx_primary = old_idx_primary;
-			vty_out(vty, "%% Failed announcing primary '%s' to peer%s", argv[0], VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-		vty_out(vty, "%% Local address '%s' announced as primary to the peer on the active socket%s", argv[0], VTY_NEWLINE);
-	}
-	return CMD_SUCCESS;
-}
-
-DEFUN_ATTR(asp_no_local_ip, asp_no_local_ip_cmd,
-	   "no local-ip " VTY_IPV46_CMD,
-	   NO_STR "Specify Local IP Address from which to contact ASP\n"
-	   "Local IPv4 Address from which to contact of ASP\n"
-	   "Local IPv6 Address from which to contact of ASP\n",
-	   CMD_ATTR_NODE_EXIT)
-{
-	struct osmo_ss7_asp *asp = vty->index;
-	int idx = ss7_asp_peer_find_host(&asp->cfg.local, argv[0]);
-	int rc;
-
-	if (idx < 0) {
-		vty_out(vty, "%% Local address '%s' not found in set%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	if (ss7_asp_is_started(asp)) {
-		if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA) {
-			if ((rc = ss7_asp_apply_drop_local_address(asp, idx)) < 0) {
-				vty_out(vty, "%% Failed removing local address '%s' from existing socket%s", argv[0], VTY_NEWLINE);
-				return CMD_WARNING;
-			}
-			vty_out(vty, "%% Local address '%s' removed from active socket connection%s", argv[0], VTY_NEWLINE);
-		}
-	}
-
-	if (osmo_ss7_asp_peer_del_host(&asp->cfg.local, argv[0]) != 0) {
-		vty_out(vty, "%% Failed deleting local address '%s' from set%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
 	return CMD_SUCCESS;
 }
 
@@ -884,50 +732,11 @@ DEFUN_ATTR(asp_remote_ip, asp_remote_ip_cmd,
 {
 	struct osmo_ss7_asp *asp = vty->index;
 	bool is_primary = argc > 1;
-	int old_idx_primary = asp->cfg.remote.idx_primary;
-	int rc;
 
 	if (osmo_ss7_asp_peer_add_host2(&asp->cfg.remote, asp, argv[0], is_primary) != 0) {
 		vty_out(vty, "%% Failed adding host '%s' to set%s", argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
-
-	if (!ss7_asp_is_started(asp))
-		return CMD_SUCCESS;
-	if (asp->cfg.proto == OSMO_SS7_ASP_PROT_IPA)
-		return CMD_SUCCESS;
-
-	/* The SCTP socket is already created, dynamically apply the new primary if it changed: */
-	if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA && ss7_asp_is_started(asp)) {
-		if ((rc = ss7_asp_apply_primary_address(asp)) < 0) {
-			/* Failed, rollback changes: */
-			asp->cfg.remote.idx_primary = old_idx_primary;
-			vty_out(vty, "%% Failed applying primary on host '%s'%s", argv[0], VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-	}
-	return CMD_SUCCESS;
-}
-
-DEFUN_ATTR(asp_no_remote_ip, asp_no_remote_ip_cmd,
-	   "no remote-ip " VTY_IPV46_CMD,
-	   NO_STR  "Specify Remote IP Address of ASP\n"
-	   "Remote IPv4 Address of ASP\n"
-	   "Remote IPv6 Address of ASP\n",
-	   CMD_ATTR_NODE_EXIT)
-{
-	struct osmo_ss7_asp *asp = vty->index;
-	int idx = ss7_asp_peer_find_host(&asp->cfg.remote, argv[0]);
-
-	if (idx < 0) {
-		vty_out(vty, "%% Remote address '%s' not found in set%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	if (osmo_ss7_asp_peer_del_host(&asp->cfg.remote, argv[0]) != 0) {
-		vty_out(vty, "%% Failed deleting remote address '%s' from set%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
-	}
 	return CMD_SUCCESS;
 }
 
@@ -1182,219 +991,15 @@ static char *as_list_for_asp(const struct osmo_ss7_asp *asp, char *buf, size_t b
 	return buf;
 }
 
-/* Similar to osmo_sock_multiaddr_get_name_buf(), but aimed at listening sockets (only local part): */
-static char *get_sockname_buf(char *buf, size_t buf_len, int fd, int proto, bool local)
-{
-	char hostbuf[OSMO_SOCK_MAX_ADDRS][INET6_ADDRSTRLEN];
-	size_t num_hostbuf = ARRAY_SIZE(hostbuf);
-	char portbuf[6];
-	struct osmo_strbuf sb = { .buf = buf, .len = buf_len };
-	bool need_more_bufs;
-	int rc;
-
-	rc = osmo_sock_multiaddr_get_ip_and_port(fd, proto, &hostbuf[0][0],
-						 &num_hostbuf, sizeof(hostbuf[0]),
-						 portbuf, sizeof(portbuf), local);
-	if (rc < 0)
-		return NULL;
-
-	need_more_bufs = num_hostbuf > ARRAY_SIZE(hostbuf);
-	if (need_more_bufs)
-		num_hostbuf = ARRAY_SIZE(hostbuf);
-	OSMO_STRBUF_APPEND(sb, osmo_multiaddr_ip_and_port_snprintf,
-			   &hostbuf[0][0], num_hostbuf, sizeof(hostbuf[0]), portbuf);
-	if (need_more_bufs)
-		OSMO_STRBUF_PRINTF(sb, "<need-more-bufs!>");
-
-	return buf;
-}
-
-static void show_one_asp(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	char as_buf[64];
-	char buf_loc[OSMO_SOCK_MULTIADDR_PEER_STR_MAXLEN];
-	char buf_rem[sizeof(buf_loc)];
-
-	int fd = ss7_asp_get_fd(asp);
-	if (fd > 0) {
-		int proto = ss7_asp_proto_to_ip_proto(asp->cfg.proto);
-		if (!get_sockname_buf(buf_loc, sizeof(buf_loc), fd, proto, true))
-			OSMO_STRLCPY_ARRAY(buf_loc, "<sockname-error>");
-		if (!get_sockname_buf(buf_rem, sizeof(buf_rem), fd, proto, false))
-			OSMO_STRLCPY_ARRAY(buf_rem, "<sockname-error>");
-	} else {
-		osmo_ss7_asp_peer_snprintf(buf_loc, sizeof(buf_loc), &asp->cfg.local);
-		osmo_ss7_asp_peer_snprintf(buf_rem, sizeof(buf_rem), &asp->cfg.remote);
-	}
-
-	vty_out(vty, "%-12s  %-12s  %-13s  %-4s  %-4s  %-9s  %-23s  %-23s%s",
-		asp->cfg.name,
-		as_list_for_asp(asp, as_buf, sizeof(as_buf)),
-		asp->fi ? osmo_fsm_inst_state_name(asp->fi) : "uninitialized",
-		get_value_string(osmo_ss7_asp_protocol_vals, asp->cfg.proto),
-		osmo_str_tolower(get_value_string(osmo_ss7_asp_role_names, asp->cfg.role)),
-		asp->cfg.is_server ? "server" : "client",
-		buf_loc, buf_rem,
-		VTY_NEWLINE);
-}
-
-static int show_asp(struct vty *vty, int id, const char *asp_name)
-{
-	struct osmo_ss7_instance *inst;
-	struct osmo_ss7_asp *asp = NULL;
-
-	inst = osmo_ss7_instance_find(id);
-	if (!inst) {
-		vty_out(vty, "No SS7 instance %d found%s", id, VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	if (asp_name) {
-		asp = osmo_ss7_asp_find_by_name(inst, asp_name);
-		if (!asp) {
-			vty_out(vty, "No ASP %s found%s", asp_name, VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-	}
-
-	vty_out(vty, "ASP Name      AS Name       State          Type  Role  SCTP Role  Local Addresses          Remote Addresses%s", VTY_NEWLINE);
-	vty_out(vty, "------------  ------------  -------------  ----  ----  ---------  -----------------------  -----------------------%s", VTY_NEWLINE);
-
-	if (asp) {
-		show_one_asp(vty, asp);
-		return CMD_SUCCESS;
-	}
-
-	llist_for_each_entry(asp, &inst->asp_list, list)
-		show_one_asp(vty, asp);
-	return CMD_SUCCESS;
-}
-
 DEFUN(show_cs7_asp, show_cs7_asp_cmd,
 	"show cs7 instance <0-15> asp",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP)\n")
-{
-	int id = atoi(argv[0]);
-
-	return show_asp(vty, id, NULL);
-}
-
-DEFUN(show_cs7_asp_name, show_cs7_asp_name_cmd,
-	"show cs7 instance <0-15> asp name ASP_NAME",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP)\n"
-	"Lookup ASP with a given name\n"
-	"Name of the Application Server Process (ASP)\n")
-{
-	int id = atoi(argv[0]);
-	const char *asp_name = argv[1];
-
-	return show_asp(vty, id, asp_name);
-}
-
-static void show_one_asp_remaddr_tcp(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	struct osmo_sockaddr osa = {};
-	struct tcp_info tcpi = {};
-	socklen_t len;
-	int fd, rc;
-
-	fd = ss7_asp_get_fd(asp);
-	if (fd < 0) {
-		vty_out(vty, "%-12s  %-46s  uninitialized%s", asp->cfg.name, "", VTY_NEWLINE);
-		return;
-	}
-
-	len = sizeof(osa.u.sas);
-	rc = getpeername(fd, &osa.u.sa, &len);
-
-	len = sizeof(tcpi);
-	rc = getsockopt(fd, SOL_TCP, TCP_INFO, &tcpi, &len);
-	if (rc < 0) {
-		char buf_err[128];
-		strerror_r(errno, buf_err, sizeof(buf_err));
-		vty_out(vty, "%-12s  %-46s  getsockopt(TCP_INFO) failed: %s%s",
-			asp->cfg.name, osmo_sockaddr_to_str(&osa), buf_err, VTY_NEWLINE);
-		return;
-	}
-
-	vty_out(vty, "%-12s  %-46s  TCP_%-19s  %-8u  %-8u  %-8u  %-8u%s",
-		asp->cfg.name,
-		osmo_sockaddr_to_str(&osa),
-		get_value_string(tcp_info_state_values, tcpi.tcpi_state),
-		tcpi.tcpi_snd_cwnd, tcpi.tcpi_rtt,
-		tcpi.tcpi_rto, tcpi.tcpi_pmtu,
-		VTY_NEWLINE);
-}
-
-#ifdef HAVE_LIBSCTP
-static void show_one_asp_remaddr_sctp(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	struct sctp_paddrinfo pinfo[OSMO_SOCK_MAX_ADDRS];
-	struct osmo_sockaddr osa = {};
-	size_t pinfo_cnt = ARRAY_SIZE(pinfo);
-	bool more_needed;
-	int fd, rc;
-	unsigned int i;
-
-	fd = ss7_asp_get_fd(asp);
-	if (fd < 0) {
-		vty_out(vty, "%-12s  %-46s  uninitialized%s", asp->cfg.name, "", VTY_NEWLINE);
-		return;
-	}
-
-	rc = osmo_sock_sctp_get_peer_addr_info(fd, &pinfo[0], &pinfo_cnt);
-	if (rc < 0) {
-		char buf_err[128];
-		strerror_r(errno, buf_err, sizeof(buf_err));
-		vty_out(vty, "%-12s  %-46s  getsockopt(SCTP_GET_PEER_ADDR_INFO) failed: %s%s", asp->cfg.name, "", buf_err, VTY_NEWLINE);
-		return;
-	}
-
-	more_needed = pinfo_cnt > ARRAY_SIZE(pinfo);
-	if (pinfo_cnt > ARRAY_SIZE(pinfo))
-		pinfo_cnt = ARRAY_SIZE(pinfo);
-
-	for (i = 0; i < pinfo_cnt; i++) {
-		osa.u.sas = pinfo[i].spinfo_address;
-		vty_out(vty, "%-12s  %-46s  SCTP_%-18s  %-8u  %-8u  %-8u  %-8u%s",
-			asp->cfg.name,
-			osmo_sockaddr_to_str(&osa),
-			osmo_sctp_spinfo_state_str(pinfo[i].spinfo_state),
-			pinfo[i].spinfo_cwnd, pinfo[i].spinfo_srtt,
-			pinfo[i].spinfo_rto, pinfo[i].spinfo_mtu,
-			VTY_NEWLINE);
-	}
-
-	if (more_needed)
-		vty_out(vty, "%-12s  more address buffers needed!%s", asp->cfg.name, VTY_NEWLINE);
-}
-#endif
-
-static void show_one_asp_remaddr(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	int proto = ss7_asp_proto_to_ip_proto(asp->cfg.proto);
-
-	switch (proto) {
-	case IPPROTO_TCP:
-		show_one_asp_remaddr_tcp(vty, asp);
-		break;
-#ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		show_one_asp_remaddr_sctp(vty, asp);
-		break;
-#endif
-	default:
-		vty_out(vty, "%-12s  %-46s  unknown proto %u%s", asp->cfg.name, "", proto, VTY_NEWLINE);
-		break;
-	}
-}
-
-static int show_asp_remaddr(struct vty *vty, int id, const char *asp_name)
+	SHOW_STR CS7_STR INST_STR INST_STR "Application Server Process (ASP)\n")
 {
 	struct osmo_ss7_instance *inst;
-	struct osmo_ss7_asp *asp = NULL;
+	struct osmo_ss7_asp *asp;
+	char buf[512];
+	char as_buf[64];
+	int id = atoi(argv[0]);
 
 	inst = osmo_ss7_instance_find(id);
 	if (!inst) {
@@ -1402,207 +1007,33 @@ static int show_asp_remaddr(struct vty *vty, int id, const char *asp_name)
 		return CMD_WARNING;
 	}
 
-	if (asp_name) {
-		asp = osmo_ss7_asp_find_by_name(inst, asp_name);
-		if (!asp) {
-			vty_out(vty, "No ASP %s found%s", asp_name, VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-	}
-
-	vty_out(vty, "ASP Name      Remote IP Address & Port                        State                    CWND      SRTT      RTO       MTU%s", VTY_NEWLINE);
-	vty_out(vty, "------------  ----------------------------------------------  -----------------------  --------  --------  --------  --------%s", VTY_NEWLINE);
-
-	if (asp) {
-		show_one_asp_remaddr(vty, asp);
-		return CMD_SUCCESS;
-	}
+	vty_out(vty, "                                                       Current Primary Link%s", VTY_NEWLINE);
+	vty_out(vty, "ASP Name      AS Name       State          Type  Role  Remote IPaddr & Port     SCTP Role%s", VTY_NEWLINE);
+	vty_out(vty, "------------  ------------  -------------  ----  ----  -----------------------  ---------%s", VTY_NEWLINE);
 
 	llist_for_each_entry(asp, &inst->asp_list, list) {
-		show_one_asp_remaddr(vty, asp);
+		if (asp->cfg.proto == OSMO_SS7_ASP_PROT_IPA && asp->cfg.remote.port == 0 && asp->server) {
+			struct osmo_fd *ofd = osmo_stream_srv_get_ofd(asp->server);
+			char hostbuf[64];
+			char portbuf[16];
+			osmo_sock_get_ip_and_port(ofd->fd, hostbuf, sizeof(hostbuf),
+						  portbuf, sizeof(portbuf), false);
+			snprintf(buf, sizeof(buf), "%s:%s", hostbuf, portbuf);
+		} else
+			osmo_ss7_asp_peer_snprintf(buf, sizeof(buf), &asp->cfg.remote);
+		vty_out(vty, "%-12s  %-12s  %-13s  %-4s  %-4s  %-23s  %-9s%s",
+			asp->cfg.name,
+			as_list_for_asp(asp, as_buf, sizeof(as_buf)),
+			asp->fi? osmo_fsm_inst_state_name(asp->fi) : "uninitialized",
+			get_value_string(osmo_ss7_asp_protocol_vals, asp->cfg.proto),
+			osmo_str_tolower(get_value_string(osmo_ss7_asp_role_names, asp->cfg.role)),
+			buf,
+			asp->cfg.is_server ? "server" : "client",
+			VTY_NEWLINE);
 	}
 	return CMD_SUCCESS;
 }
 
-DEFUN(show_cs7_asp_remaddr, show_cs7_asp_remaddr_cmd,
-	"show cs7 instance <0-15> asp-remaddr",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP) remote addresses information\n")
-{
-	int id = atoi(argv[0]);
-
-	return show_asp_remaddr(vty, id, NULL);
-}
-
-
-DEFUN(show_cs7_asp_remaddr_name, show_cs7_asp_remaddr_name_cmd,
-	"show cs7 instance <0-15> asp-remaddr name ASP_NAME",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP) remote addresses information\n"
-	"Lookup ASP with a given name\n"
-	"Name of the Application Server Process (ASP)\n")
-{
-	int id = atoi(argv[0]);
-	const char *asp_name = argv[1];
-
-	return show_asp_remaddr(vty, id, asp_name);
-}
-
-static void show_one_asp_assoc_status_tcp(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	struct osmo_sockaddr osa = {};
-	struct tcp_info tcpi = {};
-	socklen_t len;
-	int fd, rc;
-	int rx_pend_bytes = 0;
-
-	fd = ss7_asp_get_fd(asp);
-	if (fd < 0) {
-		vty_out(vty, "%-12s  uninitialized%s", asp->cfg.name, VTY_NEWLINE);
-		return;
-	}
-
-	len = sizeof(osa.u.sas);
-	rc = getpeername(fd, &osa.u.sa, &len);
-
-	len = sizeof(tcpi);
-	rc = getsockopt(fd, SOL_TCP, TCP_INFO, &tcpi, &len);
-	if (rc < 0) {
-		char buf_err[128];
-		strerror_r(errno, buf_err, sizeof(buf_err));
-		vty_out(vty, "%-12s  getsockopt(TCP_INFO) failed: %s%s",
-			asp->cfg.name, buf_err, VTY_NEWLINE);
-		return;
-	}
-
-	rc = ioctl(fd, FIONREAD, &rx_pend_bytes);
-
-	/* FIXME: RWND: struct tcp_info from linux/tcp.h contains more fields
-	 * than the one from netinet/tcp.h we currently use, including
-	 * "tcpi_rcv_wnd" which we could use to print RWND here. However,
-	 * linux/tcp.h seems to be missing the state defines used in
-	 * "tcp_info_state_values", so we cannot use that one instead.
-	 */
-
-	vty_out(vty, "%-12s  TCP_%-19s  %-9s  %-10s  %-8s  %-9u  %-7u  %-9u  %-46s%s",
-		asp->cfg.name,
-		get_value_string(tcp_info_state_values, tcpi.tcpi_state),
-		"-", "-", "-", tcpi.tcpi_unacked, rx_pend_bytes,
-		tcpi.tcpi_pmtu, osmo_sockaddr_to_str(&osa),
-		VTY_NEWLINE);
-}
-
-#ifdef HAVE_LIBSCTP
-static void show_one_asp_assoc_status_sctp(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	struct osmo_sockaddr osa = {};
-	struct sctp_status st;
-	socklen_t len;
-	int fd, rc;
-
-	fd = ss7_asp_get_fd(asp);
-	if (fd < 0) {
-		vty_out(vty, "%-12s  uninitialized%s", asp->cfg.name, VTY_NEWLINE);
-		return;
-	}
-
-	memset(&st, 0, sizeof(st));
-	len = sizeof(st);
-	rc = getsockopt(fd, IPPROTO_SCTP, SCTP_STATUS, &st, &len);
-	if (rc < 0) {
-		char buf_err[128];
-		strerror_r(errno, buf_err, sizeof(buf_err));
-		vty_out(vty, "%-12s  getsockopt(SCTP_STATUS) failed: %s%s", asp->cfg.name, buf_err, VTY_NEWLINE);
-		return;
-	}
-
-	osa.u.sas = st.sstat_primary.spinfo_address;
-	vty_out(vty, "%-12s  SCTP_%-18s  %-9u  %-10u  %-8u  %-9u  %-7u  %-9u  %-46s%s",
-		asp->cfg.name,
-		osmo_sctp_sstat_state_str(st.sstat_state),
-		st.sstat_instrms, st.sstat_outstrms,
-		st.sstat_rwnd, st.sstat_unackdata, st.sstat_penddata,
-		st.sstat_fragmentation_point,
-		osmo_sockaddr_to_str(&osa),
-		VTY_NEWLINE);
-}
-#endif
-
-static void show_one_asp_assoc_status(struct vty *vty, struct osmo_ss7_asp *asp)
-{
-	int proto = ss7_asp_proto_to_ip_proto(asp->cfg.proto);
-
-	switch (proto) {
-	case IPPROTO_TCP:
-		show_one_asp_assoc_status_tcp(vty, asp);
-		break;
-#ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		show_one_asp_assoc_status_sctp(vty, asp);
-		break;
-#endif
-	default:
-		vty_out(vty, "%-12s  unknown proto %u%s", asp->cfg.name, proto, VTY_NEWLINE);
-		break;
-	}
-}
-
-static int show_asp_assoc_status(struct vty *vty, int id, const char *asp_name)
-{
-	struct osmo_ss7_instance *inst;
-	struct osmo_ss7_asp *asp = NULL;
-
-	inst = osmo_ss7_instance_find(id);
-	if (!inst) {
-		vty_out(vty, "No SS7 instance %d found%s", id, VTY_NEWLINE);
-		return CMD_WARNING;
-	}
-
-	if (asp_name) {
-		asp = osmo_ss7_asp_find_by_name(inst, asp_name);
-		if (!asp) {
-			vty_out(vty, "No ASP %s found%s", asp_name, VTY_NEWLINE);
-			return CMD_WARNING;
-		}
-	}
-
-	vty_out(vty, "ASP Name      State                    InStreams  OutStreams  RWND      UnackData  PenData  FragPoint  Current Primary Remote IP Address & Port%s", VTY_NEWLINE);
-	vty_out(vty, "------------  -----------------------  ---------  ----------  --------  ---------  -------  ---------  ----------------------------------------------%s", VTY_NEWLINE);
-
-	if (asp) {
-		show_one_asp_assoc_status(vty, asp);
-		return CMD_SUCCESS;
-	}
-
-	llist_for_each_entry(asp, &inst->asp_list, list)
-		show_one_asp_assoc_status(vty, asp);
-	return CMD_SUCCESS;
-}
-
-DEFUN(show_cs7_asp_assoc_status, show_cs7_asp_assoc_status_cmd,
-	"show cs7 instance <0-15> asp-assoc-status",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP) SCTP association status\n")
-{
-	int id = atoi(argv[0]);
-
-	return show_asp_assoc_status(vty, id, NULL);
-}
-
-
-DEFUN(show_cs7_asp_assoc_status_name, show_cs7_asp_assoc_status_name_cmd,
-	"show cs7 instance <0-15> asp-assoc-status name ASP_NAME",
-	SHOW_STR CS7_STR INST_STR INST_STR
-	"Application Server Process (ASP) SCTP association information\n"
-	"Lookup ASP with a given name\n"
-	"Name of the Application Server Process (ASP)\n")
-{
-	int id = atoi(argv[0]);
-	const char *asp_name = argv[1];
-
-	return show_asp_assoc_status(vty, id, asp_name);
-}
-
 static void write_one_asp(struct vty *vty, struct osmo_ss7_asp *asp, bool show_dyn_config)
 {
 	int i;
@@ -2863,7 +2294,6 @@ static void vty_init_shared(void *ctx)
 	install_lib_element_ve(&show_cs7_user_cmd);
 	install_lib_element_ve(&show_cs7_xua_cmd);
 	install_lib_element_ve(&show_cs7_config_cmd);
-	install_lib_element(ENABLE_NODE, &cs7_asp_disconnect_cmd);
 
 	/* the mother of all VTY config nodes */
 	install_lib_element(CONFIG_NODE, &cs7_instance_cmd);
@@ -2879,18 +2309,11 @@ static void vty_init_shared(void *ctx)
 
 	install_node(&asp_node, NULL);
 	install_lib_element_ve(&show_cs7_asp_cmd);
-	install_lib_element_ve(&show_cs7_asp_name_cmd);
-	install_lib_element_ve(&show_cs7_asp_remaddr_cmd);
-	install_lib_element_ve(&show_cs7_asp_remaddr_name_cmd);
-	install_lib_element_ve(&show_cs7_asp_assoc_status_cmd);
-	install_lib_element_ve(&show_cs7_asp_assoc_status_name_cmd);
 	install_lib_element(L_CS7_NODE, &cs7_asp_cmd);
 	install_lib_element(L_CS7_NODE, &no_cs7_asp_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &cfg_description_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_remote_ip_cmd);
-	install_lib_element(L_CS7_ASP_NODE, &asp_no_remote_ip_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_local_ip_cmd);
-	install_lib_element(L_CS7_ASP_NODE, &asp_no_local_ip_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_qos_class_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_role_cmd);
 	install_lib_element(L_CS7_ASP_NODE, &asp_sctp_role_cmd);
@@ -2947,7 +2370,6 @@ void osmo_ss7_vty_init_sg(void *ctx)
 	install_lib_element(L_CS7_NODE, &cs7_xua_cmd);
 	install_lib_element(L_CS7_NODE, &no_cs7_xua_cmd);
 	install_lib_element(L_CS7_XUA_NODE, &xua_local_ip_cmd);
-	install_lib_element(L_CS7_XUA_NODE, &xua_no_local_ip_cmd);
 	install_lib_element(L_CS7_XUA_NODE, &xua_accept_dyn_asp_cmd);
 	install_lib_element(L_CS7_XUA_NODE, &xua_sctp_param_init_cmd);
 	install_lib_element(L_CS7_XUA_NODE, &xua_no_sctp_param_init_cmd);
