@@ -52,6 +52,7 @@
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/core/linuxlist.h>
+#include <osmocom/core/linuxrbtree.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/core/timer.h>
 #include <osmocom/core/fsm.h>
@@ -72,8 +73,8 @@
 
 /* a logical connection within the SCCP instance */
 struct sccp_connection {
-	/* part of osmo_sccp_instance.list */
-	struct llist_head list;
+	/* entry in (struct sccp_instance)->connections */
+	struct rb_node node;
 	/* which instance are we part of? */
 	struct osmo_sccp_instance *inst;
 	/* which user owns us? */
@@ -233,6 +234,31 @@ static const struct osmo_prim_event_map scu_scoc_event_map[] = {
  * Timer Handling
  ***********************************************************************/
 
+/* Mostly pasted from Appendix C.4 of ITU-T Q.714 (05/2001) -- some of their descriptions are quite
+ * unintelligible out of context, for which we have our own description here. */
+const struct osmo_tdef osmo_sccp_timer_defaults[OSMO_SCCP_TIMERS_LEN] = {
+	{ .T = OSMO_SCCP_TIMER_CONN_EST,	.default_val = 1*60,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting for connection confirm message, 1 to 2 minutes" },
+	{ .T = OSMO_SCCP_TIMER_IAS,		.default_val = 7*60,	.unit = OSMO_TDEF_S,
+	  .desc = "Send keep-alive: on an idle connection, delay before sending an Idle Timer message, 5 to 10 minutes" }, /* RFC 3868 Ch. 8. */
+	{ .T = OSMO_SCCP_TIMER_IAR,		.default_val = 15*60,	.unit = OSMO_TDEF_S,
+	  .desc = "Receive keep-alive: on an idle connection, delay until considering a connection as stale, 11 to 21 minutes" }, /* RFC 3868 Ch. 8. */
+	{ .T = OSMO_SCCP_TIMER_REL,		.default_val = 10,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting for release complete message, 10 to 20 seconds" },
+	{ .T = OSMO_SCCP_TIMER_REPEAT_REL,	.default_val = 10,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting for release complete message; or to repeat sending released message after the initial expiry, 10 to 20 seconds" },
+	{ .T = OSMO_SCCP_TIMER_INT,		.default_val = 1*60,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting for release complete message; or to release connection resources, freeze the LRN and "
+		  "alert a maintenance function after the initial expiry, extending to 1 minute" },
+	{ .T = OSMO_SCCP_TIMER_GUARD,		.default_val = 23*60,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting to resume normal procedure for temporary connection sections during the restart procedure, 23 to 25 minutes" },
+	{ .T = OSMO_SCCP_TIMER_RESET,		.default_val = 10,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting to release temporary connection section or alert maintenance function after reset request message is sent, 10 to 20 seconds" },
+	{ .T = OSMO_SCCP_TIMER_REASSEMBLY,	.default_val = 10,	.unit = OSMO_TDEF_S,
+	  .desc = "Waiting to receive all the segments of the remaining segments, single segmented message after receiving the first segment, 10 to 20 seconds" },
+	{}
+};
+
 /* Appendix C.4 of ITU-T Q.714 */
 const struct value_string osmo_sccp_timer_names[] = {
 	{ OSMO_SCCP_TIMER_CONN_EST, "conn_est" },
@@ -247,93 +273,16 @@ const struct value_string osmo_sccp_timer_names[] = {
 	{}
 };
 
-/* Mostly pasted from Appendix C.4 of ITU-T Q.714 (05/2001) -- some of their descriptions are quite
- * unintelligible out of context, for which we have our own description here. */
-const struct value_string osmo_sccp_timer_descriptions[] = {
-	{ OSMO_SCCP_TIMER_CONN_EST,
-	  "Waiting for connection confirm message, 1 to 2 minutes" },
-	{ OSMO_SCCP_TIMER_IAS,
-	  "Send keep-alive: on an idle connection, delay before sending an Idle Timer message,"
-	  " 5 to 10 minutes" },
-	{ OSMO_SCCP_TIMER_IAR,
-	  "Receive keep-alive: on an idle connection, delay until considering a connection as stale,"
-	  " 11 to 21 minutes" },
-	{ OSMO_SCCP_TIMER_REL,
-	  "Waiting for release complete message, 10 to 20 seconds" },
-	{ OSMO_SCCP_TIMER_REPEAT_REL,
-	  "Waiting for release complete message; or to repeat sending released message after the initial"
-	  " expiry, 10 to 20 seconds" },
-	{ OSMO_SCCP_TIMER_INT,
-	  "Waiting for release complete message; or to release connection resources, freeze the LRN and"
-	  " alert a maintenance function after the initial expiry, extending to 1 minute" },
-	{ OSMO_SCCP_TIMER_GUARD,
-	  "Waiting to resume normal procedure for temporary connection sections during the restart"
-	  " procedure, 23 to 25 minutes" },
-	{ OSMO_SCCP_TIMER_RESET,
-	  "Waiting to release temporary connection section or alert maintenance function after reset"
-	  " request message is sent, 10 to 20 seconds" },
-	{ OSMO_SCCP_TIMER_REASSEMBLY,
-	  "Waiting to receive all the segments of the remaining segments, single segmented message after"
-	  " receiving the first segment, 10 to 20 seconds" },
-	{}
-};
-
-/* Appendix C.4 of ITU-T Q.714 */
-const struct osmo_sccp_timer_val osmo_sccp_timer_defaults[] = {
-	[OSMO_SCCP_TIMER_CONN_EST]	= { .s =  1 * 60, },
-	[OSMO_SCCP_TIMER_IAS]		= { .s =  7 * 60, }, /* RFC 3868 Ch. 8. */
-	[OSMO_SCCP_TIMER_IAR]		= { .s = 15 * 60, }, /* RFC 3868 Ch. 8. */
-	[OSMO_SCCP_TIMER_REL]		= { .s =      10, },
-	[OSMO_SCCP_TIMER_REPEAT_REL]	= { .s =      10, },
-	[OSMO_SCCP_TIMER_INT]		= { .s =  1 * 60, },
-	[OSMO_SCCP_TIMER_GUARD]		= { .s = 23 * 60, },
-	[OSMO_SCCP_TIMER_RESET]		= { .s =      10, },
-	[OSMO_SCCP_TIMER_REASSEMBLY]	= { .s =      10, },
-};
-
-osmo_static_assert(ARRAY_SIZE(osmo_sccp_timer_defaults) == OSMO_SCCP_TIMERS_COUNT,
+osmo_static_assert(ARRAY_SIZE(osmo_sccp_timer_defaults) == (OSMO_SCCP_TIMERS_LEN) &&
+		   ARRAY_SIZE(osmo_sccp_timer_names) == (OSMO_SCCP_TIMERS_LEN),
 		   assert_osmo_sccp_timers_count);
 
-/* Look up an SCCP timer value as configured in the osmo_ss7_instance.
- * If no user defined value is set, return the global default from osmo_sccp_timer_defaults instead.
- * However, if default_if_unset is passed false, return NULL in case there is no user defined setting, or
- * in case it is identical to the global default. */
-const struct osmo_sccp_timer_val *osmo_sccp_timer_get(const struct osmo_sccp_instance *inst,
-						      enum osmo_sccp_timer timer,
-						      bool default_if_unset)
-{
-	const struct osmo_sccp_timer_val *val;
-	const struct osmo_sccp_timer_val *def;
-
-	OSMO_ASSERT(timer >= 0
-		    && timer < ARRAY_SIZE(inst->timers)
-		    && timer < ARRAY_SIZE(osmo_sccp_timer_defaults));
-
-	val = &inst->timers[timer];
-	def = &osmo_sccp_timer_defaults[timer];
-
-	/* Assert that all timer definitions have a sane global default */
-	OSMO_ASSERT(def->s || def->us);
-
-	if (val->s || val->us) {
-		if (!default_if_unset && val->s == def->s && val->us == def->us)
-			return NULL;
-		return val;
-	}
-
-	if (!default_if_unset)
-		return NULL;
-
-	/* If unset, use the global default. */
-	return def;
-}
-
 static void sccp_timer_schedule(const struct sccp_connection *conn,
 				struct osmo_timer_list *timer,
 				enum osmo_sccp_timer timer_name)
 {
-	const struct osmo_sccp_timer_val *val = osmo_sccp_timer_get(conn->inst, timer_name, true);
-	osmo_timer_schedule(timer, val->s, val->us);
+	const unsigned long val_sec = osmo_tdef_get(conn->inst->tdefs, timer_name, OSMO_TDEF_S, -1);
+	osmo_timer_schedule(timer, val_sec, 0);
 }
 
 /* T(ias) has expired, send a COIT message to the peer */
@@ -452,14 +401,47 @@ static void conn_destroy(struct sccp_connection *conn);
 static struct sccp_connection *conn_find_by_id(const struct osmo_sccp_instance *inst, uint32_t id)
 {
 	struct sccp_connection *conn;
+	const struct rb_node *node = inst->connections.rb_node;
 
-	llist_for_each_entry(conn, &inst->connections, list) {
-		if (conn->conn_id == id)
+	while (node) {
+		conn = container_of(node, struct sccp_connection, node);
+		if (id < conn->conn_id)
+			node = node->rb_left;
+		else if (id > conn->conn_id)
+			node = node->rb_right;
+		else
 			return conn;
 	}
 	return NULL;
 }
 
+static int conn_add_node(struct osmo_sccp_instance *inst, struct sccp_connection *conn)
+{
+	struct rb_node **n = &(inst->connections.rb_node);
+	struct rb_node *parent = NULL;
+
+	while (*n) {
+		struct sccp_connection *it;
+
+		it = container_of(*n, struct sccp_connection, node);
+
+		parent = *n;
+		if (conn->conn_id < it->conn_id) {
+			n = &((*n)->rb_left);
+		} else if (conn->conn_id > it->conn_id) {
+			n = &((*n)->rb_right);
+		} else {
+			LOGP(DLSCCP, LOGL_ERROR,
+			     "Trying to reserve already reserved conn_id %u\n", conn->conn_id);
+			return -EEXIST;
+		}
+	}
+
+	rb_link_node(&conn->node, parent, n);
+	rb_insert_color(&conn->node, &inst->connections);
+	return 0;
+}
+
 bool osmo_sccp_conn_id_exists(const struct osmo_sccp_instance *inst, uint32_t id)
 {
 	return conn_find_by_id(inst, id) ? true : false;
@@ -477,7 +459,10 @@ static struct sccp_connection *conn_create_id(struct osmo_sccp_user *user, uint3
 	conn->inst = user->inst;
 	conn->user = user;
 
-	llist_add_tail(&conn->list, &user->inst->connections);
+	if (conn_add_node(user->inst, conn) < 0) {
+		talloc_free(conn);
+		return NULL;
+	}
 
 	INIT_TIMER(&conn->t_conn, conn_tmr_cb, conn);
 	INIT_TIMER(&conn->t_ias, tx_inact_tmr_cb, conn);
@@ -494,7 +479,7 @@ static struct sccp_connection *conn_create_id(struct osmo_sccp_user *user, uint3
 	conn->fi = osmo_fsm_inst_alloc(&sccp_scoc_fsm, conn, conn,
 					LOGL_DEBUG, name);
 	if (!conn->fi) {
-		llist_del(&conn->list);
+		rb_erase(&conn->node, &user->inst->connections);
 		talloc_free(conn);
 		return NULL;
 	}
@@ -502,18 +487,45 @@ static struct sccp_connection *conn_create_id(struct osmo_sccp_user *user, uint3
 	return conn;
 }
 
+/* Return an unused SCCP connection ID.
+ * Callers should check the returned value: on negative return value, there are no unused IDs available.
+ * \param[in] sccp  The SCCP instance to determine a new connection ID for.
+ * \return unused ID on success (range [0x0, 0x00fffffe]) or negative on elapsed max_attempts without an unused id (<0).
+ */
+int osmo_sccp_instance_next_conn_id(struct osmo_sccp_instance *sccp)
+{
+	int max_attempts = 0x00FFFFFE;
+
+	/* SUA: RFC3868 sec 3.10.4:
+	*    The source reference number is a 4 octet long integer.
+	*    This is allocated by the source SUA instance.
+	* M3UA/SCCP: ITU-T Q.713 sec 3.3:
+	*    The "source local reference" parameter field is a three-octet field containing a
+	*    reference number which is generated and used by the local node to identify the
+	*    connection section after the connection section is set up.
+	*    The coding "all ones" is reserved for future use.
+	* Hence, as we currently use the connection ID also as local reference,
+	* let's simply use 24 bit ids to fit all link types (excluding 0x00ffffff).
+	*/
+	while (OSMO_LIKELY((max_attempts--) > 0)) {
+		/* Optimized modulo operation (% 0x00FFFFFE) using bitwise AND plus CMP: */
+		sccp->next_id = (sccp->next_id + 1) & 0x00FFFFFF;
+		if (OSMO_UNLIKELY(sccp->next_id == 0x00FFFFFF))
+			sccp->next_id = 0;
+
+		if (!conn_find_by_id(sccp, sccp->next_id))
+			return sccp->next_id;
+	}
+
+	return -1;
+}
+
 /* Search for next free connection ID and allocate conn */
 static struct sccp_connection *conn_create(struct osmo_sccp_user *user)
 {
-	uint32_t conn_id;
-
-	do {
-		/* modulo 2^24 as we currently use the connection ID also as local
-		 * reference, and that is limited to 24 bits */
-		user->inst->next_id = (user->inst->next_id + 1) %  (1 << 24);
-		conn_id = user->inst->next_id;
-	} while (conn_find_by_id(user->inst, conn_id));
-
+	int conn_id = osmo_sccp_instance_next_conn_id(user->inst);
+	if (conn_id < 0)
+		return NULL;
 	return conn_create_id(user, conn_id);
 }
 
@@ -534,7 +546,7 @@ static void conn_destroy(struct sccp_connection *conn)
 	conn_stop_connect_timer(conn);
 	conn_stop_inact_timers(conn);
 	conn_stop_release_timers(conn);
-	llist_del(&conn->list);
+	rb_erase(&conn->node, &conn->inst->connections);
 
 	osmo_fsm_inst_term(conn->fi, OSMO_FSM_TERM_REQUEST, NULL);
 
@@ -628,11 +640,12 @@ static bool xua_drop_data_check_drop(const struct osmo_scu_prim *prim, unsigned
 static bool xua_opt_data_cache_keep(struct sccp_connection *conn, const struct osmo_scu_prim *prim, int msg_type)
 {
 	uint8_t *buf;
+	uint32_t max_optional_data = conn->inst->max_optional_data;
 
 	if (xua_drop_data_check_drop(prim, SCCP_MAX_DATA, "cache overrun"))
 		return false;
 
-	if (msgb_l2len(prim->oph.msg) > SCCP_MAX_OPTIONAL_DATA) {
+	if (msgb_l2len(prim->oph.msg) > max_optional_data) {
 		if (conn->opt_data_cache) {
 			/* Caching optional data, but there already is optional data occupying the cache: */
 			LOGP(DLSCCP, LOGL_ERROR, "replacing unsent %u bytes of optional data cache with %s optional data\n",
@@ -655,6 +668,8 @@ static bool xua_opt_data_cache_keep(struct sccp_connection *conn, const struct o
 /* Check optional Data size limit, cache if necessary, return indication whether original opt data should be sent */
 static bool xua_opt_data_length_lim(struct sccp_connection *conn, const struct osmo_scu_prim *prim, int msg_type)
 {
+	uint32_t max_optional_data = conn->inst->max_optional_data;
+
 	if (!(prim && msgb_l2(prim->oph.msg) && msgb_l2len(prim->oph.msg)))
 		return false;
 
@@ -663,7 +678,7 @@ static bool xua_opt_data_length_lim(struct sccp_connection *conn, const struct o
 	case SUA_CO_COAK: /* §4.3 Connection confirm (CC) */
 		return xua_opt_data_cache_keep(conn, prim, msg_type);
 	case SUA_CO_COREF: /* §4.4 Connection refused (CREF) */
-		if (xua_drop_data_check_drop(prim, SCCP_MAX_OPTIONAL_DATA, "over ITU-T Rec. Q.713 §4.4 limit")) {
+		if (xua_drop_data_check_drop(prim, max_optional_data, "over ITU-T Rec. Q.713 §4.4 limit")) {
 			/* From the state diagrams in ITU-T Rec Q.714, there's no way to send DT1 neither before nor after CREF
 			 * at this point, so the only option we have is to drop optional data:
 			 * see Figure C.3 / Q.714 (sheet 2 of 6) */
@@ -671,7 +686,7 @@ static bool xua_opt_data_length_lim(struct sccp_connection *conn, const struct o
 		}
 		break;
 	case SUA_CO_RELRE: /* §4.5 Released (RLSD) */
-		if (msgb_l2len(prim->oph.msg) > SCCP_MAX_OPTIONAL_DATA) {
+		if (msgb_l2len(prim->oph.msg) > max_optional_data) {
 			if (xua_drop_data_check_drop(prim, SCCP_MAX_DATA, "protocol error"))
 				return false;
 			/* There's no need to cache the optional data since the connection is still active at this point:
@@ -1420,6 +1435,8 @@ static const struct osmo_fsm_state sccp_scoc_states[] = {
 				 S(SCOC_E_CONN_TMR_EXP) |
 				 S(SCOC_E_RCOC_CREF_IND) |
 				 S(SCOC_E_RCOC_ROUT_FAIL_IND),
+		.out_state_mask = S(S_IDLE) |
+				  S(S_DISCONN_PEND),
 	},
 };
 
@@ -1923,10 +1940,12 @@ int osmo_sccp_user_sap_down(struct osmo_sccp_user *scu, struct osmo_prim_hdr *op
 
 void sccp_scoc_flush_connections(struct osmo_sccp_instance *inst)
 {
-	struct sccp_connection *conn, *conn2;
-
-	llist_for_each_entry_safe(conn, conn2, &inst->connections, list)
+	struct rb_node *node;
+	while ((node = rb_first(&inst->connections))) {
+		struct sccp_connection *conn = container_of(node, struct sccp_connection, node);
 		conn_destroy(conn);
+	}
+
 }
 
 #include <osmocom/vty/vty.h>
@@ -1959,11 +1978,14 @@ static void vty_show_connection(struct vty *vty, struct sccp_connection *conn)
 void sccp_scoc_show_connections(struct vty *vty, struct osmo_sccp_instance *inst)
 {
 	struct sccp_connection *conn;
+	struct rb_node *node;
 
 	vty_out(vty, "I Local              Conn.            Remote            %s", VTY_NEWLINE);
 	vty_out(vty, "O Ref    SSN PC      State            Ref    SSN PC     %s", VTY_NEWLINE);
 	vty_out(vty, "- ------ --- ------- ---------------- ------ --- -------%s", VTY_NEWLINE);
 
-	llist_for_each_entry(conn, &inst->connections, list)
+	for (node = rb_first(&inst->connections); node; node = rb_next(node)) {
+		conn = container_of(node, struct sccp_connection, node);
 		vty_show_connection(vty, conn);
+	}
 }
