@@ -35,6 +35,7 @@
 #include <osmocom/sigtran/mtp_sap.h>
 #include <osmocom/sigtran/protocol/mtp.h>
 #include <osmocom/sigtran/sccp_helpers.h>
+#include <osmocom/sccp/sccp_types.h>
 
 #include "sccp_internal.h"
 #include "xua_internal.h"
@@ -229,13 +230,17 @@ osmo_sccp_instance_create(struct osmo_ss7_instance *ss7, void *priv)
 
 	inst->ss7 = ss7;
 	inst->priv = priv;
-	INIT_LLIST_HEAD(&inst->connections);
 	INIT_LLIST_HEAD(&inst->users);
 
 	inst->ss7_user.inst = ss7;
 	inst->ss7_user.name = "SCCP";
 	inst->ss7_user.prim_cb = mtp_user_prim_cb;
 	inst->ss7_user.priv = inst;
+	inst->max_optional_data = SCCP_MAX_OPTIONAL_DATA;
+
+	inst->tdefs = talloc_memdup(inst, osmo_sccp_timer_defaults,
+				    sizeof(osmo_sccp_timer_defaults));
+	osmo_tdefs_reset(inst->tdefs);
 
 	rc = sccp_scmg_init(inst);
 	if (rc < 0) {
@@ -265,6 +270,16 @@ void osmo_sccp_instance_destroy(struct osmo_sccp_instance *inst)
 	talloc_free(inst);
 }
 
+void osmo_sccp_set_priv(struct osmo_sccp_instance *sccp, void *priv)
+{
+	sccp->priv = priv;
+}
+
+void *osmo_sccp_get_priv(struct osmo_sccp_instance *sccp)
+{
+	return sccp->priv;
+}
+
 /*! \brief derive a basic local SCCP-Address from a given SCCP instance.
  *  \param[out] dest_addr pointer to output address memory
  *  \param[in] inst SCCP instance
@@ -619,25 +634,59 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 			if (!asp)
 				goto out_rt;
 			asp_created = true;
+			/* Ensure that the ASP we use is set to sctp-role client. */
+			asp->cfg.is_server = false;
+			/* Ensure that the ASP we use is set to role ASP. */
+			asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
 			if (default_local_ip)
 				osmo_ss7_asp_peer_set_hosts(&asp->cfg.local, asp, &default_local_ip, 1);
 			if (default_remote_ip)
 				osmo_ss7_asp_peer_set_hosts(&asp->cfg.remote, asp, &default_remote_ip, 1);
 			/* Make sure proper defaults are applied if app didn't provide specific default values */
-			osmo_ss7_asp_set_default_peer_hosts(asp);
+			ss7_asp_set_default_peer_hosts(asp);
 			asp->simple_client_allocated = true;
 		}
 
 		osmo_ss7_as_add_asp(as, asp->cfg.name);
 	}
 
-	/* Ensure that the ASP we use is set to client mode. */
-	asp->cfg.is_server = false;
+	/* Extra sanity checks if the ASP asp-clnt-* was pre-configured over VTY: */
+	if (!asp->simple_client_allocated) {
+		/* Forbid ASPs defined through VTY that are not entirely
+		 * configured. "role" and "sctp-role" must be explicitly provided:
+		 */
+		if (!asp->cfg.role_set_by_vty) {
+			LOGP(DLSCCP, LOGL_ERROR,
+			     "%s: ASP %s defined in VTY but 'role' was not set there, please set it.\n",
+			     name, asp->cfg.name);
+			goto out_asp;
+		}
+		if (!asp->cfg.sctp_role_set_by_vty) {
+			LOGP(DLSCCP, LOGL_ERROR,
+			     "%s: ASP %s defined in VTY but 'sctp-role' was not set there, please set it.\n",
+			     name, asp->cfg.name);
+			goto out_asp;
+		}
 
-	/* Make sure that the role of this ASP is set to ASP unless the user
-	 * made a concious decision about the role via the VTY */
-	if (!asp->cfg.role_set_by_vty)
-		asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
+		/* If ASP was configured through VTY it may be explicitly configured as
+		 * SCTP server. It may be a bit confusing since this function is to create
+		 * a "SCCP simple client", but this allows users of this API such as
+		 * osmo-hnbgw to support SCTP-role server if properly configured through VTY.
+		*/
+		if (asp->cfg.is_server) {
+			struct osmo_xua_server *xs;
+			LOGP(DLSCCP, LOGL_NOTICE,
+			     "%s: Requesting an SCCP simple client on ASP %s configured with 'sctp-role server'\n",
+			     name, asp->cfg.name);
+			xs = osmo_ss7_xua_server_find(ss7, prot, asp->cfg.local.port);
+			if (!xs) {
+				LOGP(DLSCCP, LOGL_ERROR, "%s: Requesting an SCCP simple client on ASP %s configured "
+				     "with 'sctp-role server' but no matching xUA server was configured!\n",
+				     name, asp->cfg.name);
+				goto out_asp;
+			}
+		}
+	}
 
 	/* Restart ASP */
 	if (prot != OSMO_SS7_ASP_PROT_IPA)
@@ -816,6 +865,24 @@ out_strings:
 	return NULL;
 }
 
+/*! Adjust the upper bound for the optional data length (the payload) for CR, CC, CREF and RLSD messages.
+ * For any Optional Data part larger than this value in octets, send CR, CC, CREF and RLSD messages without any payload,
+ * and send the data payload in a separate Data Form 1 message. ITU-T Q.713 sections 4.2 thru 4.5 define a limit of 130
+ * bytes for the 'Data' parameter. This limit can be adjusted here. May be useful for interop with nonstandard SCCP
+ * peers.
+ * \param[in] sccp  SCCP instance to reconfigure.
+ * \param[in] val  Number of bytes to set as upper bound for the optional data length, or pass a negative value to set
+ *                 the standard value of SCCP_MAX_OPTIONAL_DATA == 130, which conforms to ITU-T Q.713.
+ */
+void osmo_sccp_set_max_optional_data(struct osmo_sccp_instance *inst, int val)
+{
+	if (!inst)
+		return;
+	if (val < 0)
+		val = SCCP_MAX_OPTIONAL_DATA;
+	inst->max_optional_data = val;
+}
+
 /*! \brief get the SS7 instance that is related to the given SCCP instance
  *  \param[in] sccp SCCP instance
  *  \returns SS7 instance; NULL if sccp was NULL */
