@@ -258,7 +258,7 @@ static struct xua_msg *sua_gen_error_msg(uint32_t err_code, struct msgb *msg)
 }
 
 /***********************************************************************
- * Transmitting SUA messsages to SCTP
+ * Transmitting SUA messages to SCTP
  ***********************************************************************/
 
 static struct msgb *sua_to_msg(struct xua_msg *xua)
@@ -316,7 +316,7 @@ int sua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 		return -1;
 
 	/* send the msg to the AS for transmission.  The AS FSM might
-	 * (depending on its state) enqueue it before trnsmission */
+	 * (depending on its state) enqueue it before transmission */
 	rc = osmo_fsm_inst_dispatch(as->fi, XUA_AS_E_TRANSFER_REQ, msg);
 	if (rc < 0)
 		msgb_free(msg);
@@ -324,7 +324,7 @@ int sua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 }
 
 /***********************************************************************
- * Receiving SUA messsages from SCTP
+ * Receiving SUA messages from SCTP
  ***********************************************************************/
 
 /*! \brief Decode SUA Global Title according to RFC3868 Section 3.10.2.3
@@ -427,7 +427,7 @@ int sua_addr_parse_part(struct osmo_sccp_addr *out,
 	 * probably want to have a separate general parsing function storing
 	 * the subparts in xua_msg_part. But before we do, we should find more
 	 * users of this subpart parsing and be aware of the performance
-	 * tradeoff.
+	 * trade-off.
 	 */
 
 	while (pos + sizeof(*par) < param->len) {
@@ -619,7 +619,7 @@ static int sua_rx_mgmt_ntfy(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	if (ntfy.info_string)
 		talloc_free(ntfy.info_string);
 
-	/* TODO: should we report this soemwhere? */
+	/* TODO: should we report this somewhere? */
 	return 0;
 }
 
@@ -680,7 +680,7 @@ int sua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 	OSMO_ASSERT(asp->cfg.proto == OSMO_SS7_ASP_PROT_SUA);
 
 	/* caller owns msg memory, we shall neither free it here nor
-	 * keep references beyon the execution of this function and its
+	 * keep references beyond the execution of this function and its
 	 * callees. */
 
 	if (!asp->inst->sccp) {
@@ -783,7 +783,7 @@ static struct xua_msg *sua_encode_duna(const uint32_t *rctx, unsigned int num_rc
 
 	xua->hdr = XUA_HDR(SUA_MSGC_SNM, SUA_SNM_DUNA);
 	xua->hdr.version = SUA_VERSION;
-	if (rctx)
+	if (rctx && num_rctx)
 		xua_msg_add_data(xua, SUA_IEI_ROUTE_CTX, num_rctx * 4, (const uint8_t *)rctx);
 
 	xua_msg_add_data(xua, SUA_IEI_AFFECTED_PC, num_aff_pc * 4, (const uint8_t *) aff_pc);
@@ -838,7 +838,7 @@ static struct xua_msg *sua_encode_dupu(const uint32_t *rctx, unsigned int num_rc
 
 	xua->hdr = XUA_HDR(SUA_MSGC_SNM, SUA_SNM_DUNA);
 	xua->hdr.version = SUA_VERSION;
-	if (rctx)
+	if (rctx && num_rctx)
 		xua_msg_add_data(xua, SUA_IEI_ROUTE_CTX, num_rctx * sizeof(*rctx), (const uint8_t *)rctx);
 
 	xua_msg_add_u32(xua, SUA_IEI_AFFECTED_PC, dpc);
@@ -875,6 +875,35 @@ void sua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsign
 	sua_tx_xua_asp(asp, xua);
 }
 
+
+/*! Transmit SSNM SCON message indicating congestion
+ *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
+ *  \param[in] rctx array of Routing Contexts in network byte order.
+ *  \param[in] num_rctx number of rctx
+ *  \param[in] aff_pc array of 'Affected Point Code' in network byte order.
+ *  \param[in] num_aff_pc number of aff_pc
+ *  \param[in] ssn optional SSN (can be NULL)
+ *  \param[in] cong_level optional congestion level (can be NULL)
+ *  \param[in] info_string optional information string (can be NULL). */
+void sua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+			   const uint32_t *aff_pc, unsigned int num_aff_pc, const uint32_t *ssn,
+			   const uint8_t cong_level, const char *info_string)
+{
+	struct xua_msg *xua = xua_msg_alloc();
+
+	xua->hdr = XUA_HDR(SUA_MSGC_SNM, SUA_SNM_SCON);
+	xua->hdr.version = SUA_VERSION;
+	if (rctx && num_rctx)
+		xua_msg_add_data(xua, SUA_IEI_ROUTE_CTX, num_rctx * sizeof(*rctx), (const uint8_t *)rctx);
+	xua_msg_add_data(xua, SUA_IEI_AFFECTED_PC, num_aff_pc * sizeof(*aff_pc), (const uint8_t *) aff_pc);
+	if (ssn)
+		xua_msg_add_u32(xua, SUA_IEI_SSN, *ssn);
+	xua_msg_add_u32(xua, SUA_IEI_CONG_LEVEL, cong_level);
+	if (info_string)
+		xua_msg_add_data(xua, SUA_IEI_INFO_STRING, strlen(info_string)+1, (const uint8_t *) info_string);
+}
+
+
 /*! Transmit SSNM DUPU message indicating user unavailability.
  *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
  *  \param[in] rctx array of Routing Contexts in network byte order.
