@@ -65,8 +65,6 @@ static const struct value_string xua_asp_event_names[] = {
 	{ XUA_ASP_E_ASPSM_BEAT,		"ASPSM_BEAT" },
 	{ XUA_ASP_E_ASPSM_BEAT_ACK,	"ASPSM_BEAT_ACK" },
 
-	{ XUA_ASP_E_AS_ASSIGNED,	"AS_ASSIGNED" },
-
 	{ IPA_ASP_E_ID_RESP,		"IPA_CCM_ID_RESP" },
 	{ IPA_ASP_E_ID_GET,		"IPA_CCM_ID_GET" },
 	{ IPA_ASP_E_ID_ACK,		"IPA_CCM_ID_ACK" },
@@ -293,8 +291,6 @@ static int peer_send_error(struct osmo_fsm_inst *fi, uint32_t err_code)
 	struct xua_msg *xua = xua_msg_alloc();
 	struct msgb *msg;
 
-	LOGPFSML(fi, LOGL_ERROR, "Tx MGMT_ERR '%s'\n", get_value_string(m3ua_err_names, err_code));
-
 	xua->hdr = XUA_HDR(SUA_MSGC_MGMT, SUA_MGMT_ERR);
 	xua->hdr.version = SUA_VERSION;
 	xua_msg_add_u32(xua, SUA_IEI_ERR_CODE, err_code);
@@ -691,9 +687,6 @@ static void xua_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_ASPSM_BEAT_ACK:
 		/* FIXME: stop timer, if any */
 		break;
-	case XUA_ASP_E_AS_ASSIGNED:
-		/* Ignore, only used in IPA asps so far. */
-		break;
 	default:
 		break;
 	}
@@ -855,16 +848,16 @@ enum ipa_asp_fsm_t {
 static int get_fd_from_iafp(struct ipa_asp_fsm_priv *iafp)
 {
 	struct osmo_ss7_asp *asp = iafp->asp;
-	int fd;
+	struct osmo_fd *ofd;
 
 	if (asp->server)
-		fd = osmo_stream_srv_get_fd(asp->server);
+		ofd = osmo_stream_srv_get_ofd(asp->server);
 	else if (asp->client)
-		fd = osmo_stream_cli_get_fd(asp->client);
+		ofd = osmo_stream_cli_get_ofd(asp->client);
 	else
 		return -1;
 
-	return fd;
+	return ofd->fd;
 }
 
 /***************
@@ -1065,7 +1058,6 @@ static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ipa_asp_fsm_priv *iafp = fi->priv;
-	struct osmo_ss7_as *as;
 	int fd;
 
 	switch (event) {
@@ -1085,15 +1077,6 @@ static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		/* stop timer, if any */
 		osmo_timer_del(&iafp->pong_timer);
 		break;
-	case XUA_ASP_E_AS_ASSIGNED:
-		as = data;
-		osmo_talloc_replace_string(iafp->ipa_unit, &iafp->ipa_unit->unit_name, as->cfg.name);
-		/* Now that the AS is known, start the client side: */
-		if (iafp->role == OSMO_SS7_ASP_ROLE_ASP && fi->state == IPA_ASP_S_DOWN) {
-			LOGPFSML(fi, LOGL_NOTICE, "Bringing up ASP now once it has been assigned to an AS\n");
-			osmo_fsm_inst_dispatch(fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
-		}
-		break;
 	default:
 		break;
 	}
@@ -1192,13 +1175,12 @@ struct osmo_fsm ipa_asp_fsm = {
 	.allstate_event_mask = S(XUA_ASP_E_SCTP_COMM_DOWN_IND) |
 			       S(XUA_ASP_E_SCTP_RESTART_IND) |
 			       S(XUA_ASP_E_ASPSM_BEAT) |
-			       S(XUA_ASP_E_ASPSM_BEAT_ACK) |
-			       S(XUA_ASP_E_AS_ASSIGNED),
+			       S(XUA_ASP_E_ASPSM_BEAT_ACK),
 	.allstate_action = ipa_asp_allstate,
 };
 
 
-/*! \brief Start a new ASP finite state machine for given ASP
+/*! \brief Start a new ASP finite stae machine for given ASP
  *  \param[in] asp Application Server Process for which to start FSM
  *  \param[in] role Role (ASP, SG, IPSP) of this FSM
  *  \param[in] log_level Logging Level for ASP FSM logging
@@ -1208,9 +1190,6 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 {
 	struct osmo_fsm_inst *fi;
 	struct ipa_asp_fsm_priv *iafp;
-	struct osmo_ss7_as *as = ipa_find_as_for_asp(asp);
-	const char *unit_name;
-	bool can_start = true;
 
 	/* allocate as child of AS? */
 	fi = osmo_fsm_inst_alloc(&ipa_asp_fsm, asp, NULL, log_level, asp->cfg.name);
@@ -1220,33 +1199,16 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
 		return NULL;
 	}
-
-	if (as) {
-		unit_name = as->cfg.name;
-	} else if (asp->dyn_allocated) {
-		LOGPFSML(fi, LOGL_INFO, "Dynamic ASP is not assigned to any AS, "
-			 "using ASP name instead of AS name as ipa_unit_name\n");
-		unit_name = asp->cfg.name;
-	} else {
-		/* ASP in client mode will be brought up when this ASP is added
-		 * to an AS, see XUA_ASP_E_AS_ASSIGNED. */
-		if (role == OSMO_SS7_ASP_ROLE_ASP) {
-			LOGPFSML(fi, LOGL_NOTICE, "ASP is not assigned to any AS. ASP bring up delayed\n");
-			can_start = false;
-		}
-		unit_name = asp->cfg.name;
-	}
-
 	iafp->role = role;
 	iafp->asp = asp;
 	iafp->ipa_unit = talloc_zero(iafp, struct ipaccess_unit);
-	iafp->ipa_unit->unit_name = talloc_strdup(iafp->ipa_unit, unit_name);
+	iafp->ipa_unit->unit_name = talloc_strdup(iafp->ipa_unit, asp->cfg.name);
 	iafp->pong_timer.cb = ipa_pong_timer_cb;
 	iafp->pong_timer.data = fi;
 
 	fi->priv = iafp;
 
-	if (can_start && role == OSMO_SS7_ASP_ROLE_ASP)
+	if (role == OSMO_SS7_ASP_ROLE_ASP)
 		osmo_fsm_inst_dispatch(fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
 
 	return fi;
